# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np

from transforms3d.euler import euler2mat
from transforms3d.affines import compose
from marxs.optics import OpticalElement
from marxs.simulator import Parallel
from marxs.math.geometry import Cylinder


class Rod(OpticalElement):
    '''X-axis of the rod is the cylinder axis (x-zoom gives half-length)
    y zoom and z zoom have to be the same (no elliptical cylinders)
    '''
    outcol = 'hitrod'
    display = {'shape': 'cylinder', 'color': 'black'}

    default_geometry = Cylinder

    def process_photons(self, photons, intersect, interpos, intercoos):
        if self.outcol not in photons.colnames:
            photons[self.outcol] = False
        photons[self.outcol][intersect] = True
        photons['probability'][intersect] = 0.
        return photons


class ThreeSidedBoom(Parallel):
    '''Three-sided boom with dimensions similar to the Arcus proposal.

    Parameters
    ----------
    boom_dims : dict (optional)
        Entries in `boom_dict` will overwrite class attributes. It's meant
        to set different dimensions for the boom.
    '''
    l_longeron = 1.08 * 1e3
    l_batten = 1.6 * 1e3
    l_diag = np.sqrt(l_longeron**2 + l_batten**2)
    d_longeron = 10.2
    d_batten = 8.
    d_diag = 2.
    # If the bottom of a boom does not contribute to the absorption
    # setting start_bay to a larger number can shorten the computation time.
    start_bay = 0
    n_bays = 10
    n_sides = 3

    def __init__(self, **kwargs):
        boom_dims = kwargs.pop('boom_dimensions', {})
        for k in boom_dims:
            if hasattr(self, k):
                setattr(self, k, boom_dims[k])
            else:
                raise ValueError('{} has not attribute {}'.format(self.name, k))
        kwargs['elem_pos'] = self.pos_spec()
        kwargs['elem_class'] = Rod
        kwargs['elem_args'] = {}
        super(ThreeSidedBoom, self).__init__(**kwargs)

    def one_window(self):
        '''Positions for rods on one side of one bay

        Returns
        -------
        pos4d : list
            List of pos4d matrices
        '''
        # longeron
        zoom = [self.l_longeron / 2, self.d_longeron / 2, self.d_longeron / 2]
        trans = [self.l_longeron / 2, self.l_batten / 3**0.5, 0.]
        rot = np.eye(3)
        pos4d = [compose(trans, rot, zoom)]

        # batten
        zoom = [self.l_batten / 2, self.d_batten / 2, self.d_batten / 2]
        trans = [0., self.l_batten / 4. / 3**0.5, self.l_batten / 4]
        rot = euler2mat(np.pi / 2, - np.pi / 6, 0, 'szxz')
        pos4d.append(compose(trans, rot, zoom))

        # diagonal1
        zoom = [self.l_diag / 2., self.d_diag / 2., self.d_diag / 2.]
        trans[0] = self.l_longeron / 2.
        rot = euler2mat(np.deg2rad(90. - 34.03), -np.pi / 6, 0, 'szxz')
        pos4d.append(compose(trans, rot, zoom))

        # diagonal2
        rot = euler2mat(np.deg2rad(90. + 34.03), -np.pi / 6, 0, 'szxz')
        pos4d.append(compose(trans, rot, zoom))
        return pos4d

    def pos_spec(self):
        '''Calculate pos4d matrices for each element on the boom'''

        fullboompos4d = []
        for h in range(self.start_bay, self.n_bays):
            trans = [h * self.l_longeron, 0, 0]
            for i in range(self.n_sides):
                rot = euler2mat((i * 2) * np.pi / self.n_sides, 0, 0, 'sxyz')
                affmat = compose(trans, rot, np.ones(3))
                for p in self.one_window():
                    fullboompos4d.append(np.dot(affmat, p))

        return fullboompos4d


class FourSidedBoom(ThreeSidedBoom):
    '''Four-sided boom with dimensions according to the ARCUS proposal.

    Parameters
    ----------
    boom_dims : dict (optional)
        Entries in `boom_dict` will overwrite class attributes. It's meant
        to set different dimensions for the boom.
    '''
    l_longeron = 1350.5
    l_batten = 1308.
    l_diag = np.sqrt(l_longeron**2 + l_batten**2)
    d_longeron = 10.2
    d_batten = 8.43
    d_diag = 1.85
    n_bays = 8
    n_sides = 4

    def one_window(self):
        '''Calculate pos4d matrices for each element on the boom'''
        # longeron
        zoom = [self.l_longeron / 2, self.d_longeron / 2, self.d_longeron / 2]
        trans = [self.l_longeron / 2, self.l_batten / 2, self.l_batten / 2]
        rot = np.eye(3)
        pos4d = [compose(trans, rot, zoom)]

        # batten
        zoom = [self.l_batten / 2, self.d_batten / 2, self.d_batten / 2]
        trans = [0., self.l_batten / 2, 0]
        rot = euler2mat(np.pi / 2, 0, 0, 'syxz')
        pos4d.append(compose(trans, rot, zoom))

        # diagonal1
        zoom = [self.l_diag / 2., self.d_diag / 2., self.d_diag / 2.]
        trans = [self.l_longeron / 2., self.l_batten / 2, 0.]
        ang = np.arcsin(self.l_batten / self.l_diag)
        rot = euler2mat(ang, 0, 0, 'syxz')
        pos4d.append(compose(trans, rot, zoom))

        # diagonal2
        rot = euler2mat(- ang, 0, 0, 'syxz')
        pos4d.append(compose(trans, rot, zoom))

        return pos4d



# Licensed under GPL version 3 - see LICENSE.rst
from os.path import join as pjoin
from glob import glob
from tempfile import TemporaryDirectory

import numpy as np
import astropy.units as u
from astropy.table import Table, QTable
from astropy.utils.data import get_pkg_data_filename

from marxs.reduction import osip
from marxs.missions.arcus.arcus import defaultconf as arcusconf
from marxs.missions.arcus.utils import config
from marxs.missions.arcus import arfrmf
from marxs.optics.detector import CCDRedistNormal

import pytest


# In this example, we use a file that's included in MARXS for test purposes
filename = get_pkg_data_filename('data/ccd_redist_normal.ecsv', 'marxs.optics.tests')
tab_redist = QTable.read(filename)


def test_filename():
    n = arfrmf.filename_from_meta('arf', ARCCHAN='1111', ORDER=3)
    assert n == 'chan_all_+3.arf'

    n = arfrmf.filename_from_meta(ARCCHAN='1111', ORDER=3,
                           TRUEORD=-3, CCDORDER=5)
    assert n == 'chan_all_ccdord_+5_true_-3.fits'

    # After reading a fits file, keywords will be string.
    # Some may have been modified, so test mixture of int and string
    n = arfrmf.filename_from_meta(ARCCHAN='1111', ORDER='+3',
                           TRUEORD='-4', CCDORDER=-5)
    assert n == 'chan_all_ccdord_-5_true_-4.fits'

    # But and order is not confused by itself
    n = arfrmf.filename_from_meta(ARCCHAN='1111', ORDER=-3,
                           TRUEORD=-3, CCDORDER=-3)
    assert n == 'chan_all_ccdord_-3_true_-3.fits'


@pytest.mark.skipif('data' not in config,
                    reason='Test requires Arcus CALDB')
def test_onccd():
    '''Don't want to hardcode exact locations of CCDs in this test,
    because we might move around CCDs or aimpoints. So, instead, this test
    just checks a few generic properties and wavelengths that are wildly off
    chip.
    '''
    onccd = arfrmf.OnCCD(arcusconf)
    out = onccd(np.arange(200) * u.Angstrom, 1, '1')
    assert out.sum() > 0  # Some wavelength fall on chips
    assert out.sum() < len(out)  # but not all of them

    # chip gaps are different for different orders
    out1 = onccd(np.arange(25, 30, .001) * u.Angstrom, -5, '1')
    out2 = onccd(np.arange(25, 30, .001) * u.Angstrom, -6, '1')
    assert not np.all(out1 == out2)

    # chip gaps are at same position in m * lambda space
    out1 = onccd(np.arange(25, 30, .001) * u.Angstrom, -5, '1')
    out2 = onccd(np.arange(25, 30, .001) * u.Angstrom / 6 * 5, -6, '1')
    assert np.all(out1 == out2)

    # chip gaps are different for different opt ax
    out1 = onccd(np.arange(30, 35, .001) * u.Angstrom, -4, '1')
    out2 = onccd(np.arange(30, 35, .001) * u.Angstrom, -4, '2')
    assert not np.all(out1 == out2)


@pytest.mark.skipif('data' not in config,
                    reason='Test requires Arcus CALDB')
def test_arf_nonzero():
    '''Arcus design will evolve, so I don't want to test exact numbers here
    but if we are off my orders of magnitude that that's likely a code error.
    '''
    arf = arfrmf.mkarf([23, 24, 25, 26] * u.Angstrom, -5)
    assert np.all(arf['SPECRESP'] > 10 * u.cm**2)
    assert np.all(arf['SPECRESP'] < 400 * u.cm**2)


@pytest.mark.skipif('data' not in config,
                    reason='Test requires Arcus CALDB')
def test_arf_channels_addup():
    arfall = arfrmf.mkarf([33, 34, 35, 36] * u.Angstrom, -3)

    arflist = []
    for k in arcusconf['pos_opt_ax'].keys():
        arflist.append(arfrmf.mkarf([33, 34, 35, 36] * u.Angstrom, -3, channels=[k]))

    assert np.allclose(arfall['SPECRESP'],
                       sum([a['SPECRESP'] for a in arflist])
                       )


@pytest.mark.skipif('data' not in config,
                    reason='Test requires Arcus CALDB')
def test_mirrgrat():
    '''Test that mirr_grat works. Don't want to hardcode specific numbers
    (since they will change), but test a range here.
    '''
    aefforder = QTable.read(pjoin(config['data']['caldb_inputdata'], 'aeff',
                                     'mirr_grat.tab'), format='ascii.ecsv')
    mirrgrat = arfrmf.MirrGrat(aefforder=aefforder)
    aeff = mirrgrat([10, 20, 30] * u.Angstrom, -5)
    assert np.all(aeff < [10, 50, 100] * u.cm**2)
    assert np.all(aeff > [.1, 5., 10.] * u.cm**2)


@pytest.mark.skipif('data' not in config,
                    reason='Test requires Arcus CALDB')
def test_make_arf_osip():
    '''integration test: Generate and ARF and then apply an OSIP to it'''
    with TemporaryDirectory() as tmpdirname:
        # -5 is last, because we need that for the tests below
        for o in [-4, -6, -5]:
            arf = arfrmf.mkarf([23, 24, 25, 26] * u.Angstrom, o)
            arf = arfrmf.tagversion(arf)
            # check the default in mkarf
            assert arf.meta['ARCCHAN'] == '1111'

            basearf = pjoin(tmpdirname,
                            arfrmf.filename_from_meta('arf', **arf.meta))
            arf.write(basearf)

        osipp = osip.FixedFractionOSIP(
            0.7,
            ccd_redist=CCDRedistNormal(tab_width=tab_redist),
            filename_from_meta=arfrmf.filename_from_meta,
        )
        osipp.apply_osip_all(tmpdirname, tmpdirname, [-5],
                             filename_from_meta_kwargs={'ARCCHAN': '1111'})


        # For debugging: Keep list of all filenames in variable
        globres = glob(pjoin(tmpdirname, '*'))
        # there are three confused arfs now (to be read below)
        assert len(glob(pjoin(tmpdirname, '*-*-*'))) == 3

        # Check plots are created, but don't check content.
        # Not sure how to compare pdfs without too much work.
        assert len(glob(pjoin(tmpdirname, '*.pdf'))) > 0

        # using astropy.table here to be independent of the implementation of
        # ARF in arcus.reduction.ogip
        barf = Table.read(basearf, format='fits')

        # Using glob to get filename, to be independent of
        # filename_from_meta
        arfcenter = Table.read(glob(pjoin(tmpdirname, '*-5*-5*'))[0],
                               format='fits')
        assert np.isclose(arfcenter.meta['OSIPFAC'], 0.7)
        arfup = Table.read(glob(pjoin(tmpdirname, '*-5*-4*'))[0],
                           format='fits')
        arfdown = Table.read(glob(pjoin(tmpdirname, '*-5*-6*'))[0],
                             format='fits')
        # In this setup, some area falls between extraction regions
        assert np.all(barf['SPECRESP'] >
                      (arfcenter['SPECRESP'] +
                       arfup['SPECRESP'] +
                       arfdown['SPECRESP']))


# Licensed under GPL version 3 - see LICENSE.rst
import copy
import os

import numpy as np
import astropy.units as u
from astropy.table import Table
from astropy.utils.data import get_pkg_data_filename
import transforms3d
from scipy.interpolate import RectBivariateSpline

from marxs import optics
from marxs.missions.mitsnl.catgrating import CATL1L2Stack, l1_order_selector
from marxs.base import TagVersion
from marxs import design
from marxs.math.geometry import Cylinder
from marxs.math.utils import xyz2zxy
from marxs.missions.mitsnl.catgrating import InterpolateEfficiencyTable
from marxs.missions.athena import spo
from marxs.utils import tablerows_to_2d
from marxs import simulator
from marxs import analysis
from marxs.missions.arcus.arcus import FiltersAndQE

tagversion = TagVersion(SATELLIT='Athena', GRATING='CAT')

# SPOs reflectivity tabulated data
reflectivity = tablerows_to_2d(Table.read(
    get_pkg_data_filename('data/coated_reflectivity.csv'),
    format='ascii.ecsv'))
reflectivity_interpolator = RectBivariateSpline(reflectivity[0].to(u.keV),
                                                reflectivity[1].to(u.rad),
                                                reflectivity[2][0])


# CAT gratings tabulated data
order_selector_Si = InterpolateEfficiencyTable(Table.read(
    get_pkg_data_filename('data/Si_efficiency_5_7.dat',
                          package='marxs.missions.mitsnl'),
    format='ascii.ecsv'))
order_selector_Si.coating = 'None'
order_selector_Pt = InterpolateEfficiencyTable(Table.read(
    get_pkg_data_filename('data/SiPt_efficiency_5_7.dat',
                          package='marxs.missions.mitsnl'),
    format='ascii.ecsv'))
order_selector_Pt.coating = 'Pt'


conf = {
        'blazeang': np.deg2rad(1.6),
        'focallength': 12000.,
        'design_tilted_torus': 11.6e3,
        'alphafac': 2.2,
        'betafac': 4.4,
        # SPO scatter
        # factor 2.3545 converts from FWHM to sigma
        'perpplanescatter':  1.5 / 2.3545 * u.arcsec,
        # 2 * 0.68 converts HPD to sigma
        'inplanescatter': 7. / (2 * 0.68) * u.arcsec,
        'spo_pos4d': spo.spo_pos4d,
        'spo_geom': spo.spogeom,
        'reflectivity_interpolator': reflectivity_interpolator,

        # Due to the implementation in Rowland circle array, this is reversed
        # from what I would expect.
        'grating_size': np.array([60., 60.]),
        'grating_frame': 2.,
        'det_kwargs': {'y_range': [300, 600],
                       'd_element': [24.576 * 2 + 0.824 * 2 + 0.5,
                                     24.576 + 0.824 * 2 + 0.5],
                       'elem_class': optics.FlatDetector,
                       'elem_args': {'zoom': [1, 24.576, 12.288],
                                     'pixsize': 0.024,
                                     # orientation flips around CCDs so that det_x increases
                                     # with increasing x coordinate
                                    'orientation': np.array([[-1, 0, 0],
                                                             [0, -1, 0],
                                                             [0, 0, +1]])
                                     }},
        'gas_kwargs': {'parallel_spec': np.array([1., 0., 0., 0.]),
                       'normal_spec': np.array([0, 0, 0, 1]),
                       'opt_range': [8e3, 1.2e4],
                       'optimize_axis': 'z',
                       'radius': (np.min(spo.spogeom['inner_radius']),
                                  np.max(spo.spogeom['outer_radius'])),
                       'elem_class': CATL1L2Stack,
                       'elem_args': {'d': 2e-4,
                                     'order_selector': order_selector_Si,
                                     'l1_dims': {'bardepth': 5.7 * u.micrometer,
                                                 'period': 0.005 * u.mm,
                                                 'barwidth': 0.0005 * u.mm},
                                     'l2_dims': {'bardepth': 0.5 * u.mm,
                                                 'period': 0.966 * u.mm,
                                                 'barwidth': 0.05 * u.mm},
                                     # Set sigma to 0. Effectively disables this
                                     # factor.
                                     'qualityfactor': {'d': 200. * u.um,
                                                       'sigma': 0 * u.um},
                                     'l1_order_selector': l1_order_selector,
                                     },
                       },
    }


def add_rowland_to_conf(conf):
    conf['alpha'] = conf['alphafac'] * conf['blazeang']
    conf['beta'] = conf['betafac'] * conf['blazeang']
    R, r, pos4d = design.rowland.design_tilted_torus(conf['design_tilted_torus'],
                                                     conf['alpha'], conf['beta'])
    conf['rowland'] = design.RowlandTorus(R, r, pos4d=pos4d)
    conf['rowland'].pos4d = np.dot(xyz2zxy, conf['rowland'].pos4d)
    conf['blazemat'] = transforms3d.axangles.axangle2mat(np.array([0, 0, 1]),
                                                         -conf['blazeang'])

add_rowland_to_conf(conf)


conf_chirp = copy.copy(conf)
conf_chirp['grating_size'] = np.array([80., 160.])
conf_chirp['chirp_energy'] = 0.6
# Use of a non-integer order makes no sense physically
# but is just a numerical tool to optimize at the blaze peak
conf_chirp['chirp_order'] = -5.4


class GAS(design.rowland.CircularMeshGrid):
    def __init__(self, conf):
        gg = conf['gas_kwargs']
        # Add in / update keywords that are calculated based on other conf
        # entries
        gg['rowland'] = conf['rowland']
        # Reverse grating size for d_element to account for rotation of elementa
        # roation is dome automatic by parallel_spec, but the algorithm that makes
        # the tiling does not know about that
        gg['d_element'] = conf['grating_size'] + 2 * conf['grating_frame']
        gg['elem_args']['zoom'] = [1.,
                                   conf['grating_size'][0] / 2.,
                                   conf['grating_size'][1] / 2.]
        gg['elem_args']['orientation'] = conf['blazemat']

        super().__init__(**gg)

        # Color gratings according to the sector they are in
        for e in self.elements:
            e.display = copy.deepcopy(e.display)
            # Angle from baseline in range 0..pi
            ang = np.arctan2(e.pos4d[1, 3], e.pos4d[2, 3]) % (np.pi)
            # pick one colors
            e.display['color'] = 'rgb'[int(ang / np.pi * 3)]


class RowlandDetArray(design.rowland.RectangularGrid):
    def __init__(self, conf):
        super().__init__(conf['rowland'], **conf['det_kwargs'],
                         guess_distance=25.)


# Place an additional detector on the Rowland circle.

detcirc = optics.CircularDetector(geometry=Cylinder.from_rowland(conf['rowland'],
                                                                 width=50,
                                                                 rotation=np.pi,
                                                                 kwargs={'phi_lim':[-np.pi/2, np.pi/2]}))
detcirc.loc_coos_name = ['detcirc_phi', 'detcirc_y']
detcirc.detpix_name = ['detcircpix_x', 'detcircpix_y']
detcirc.display['opacity'] = 0.1

flatdet = optics.FlatDetector(orientation=xyz2zxy[:3, :3], zoom=[1, 1e5, 1e5])
flatdet.loc_coos_name = ['detinf_x', 'detinf_y']
flatdet.detpix_name = ['detinfpix_x', 'detinfpix_y']
flatdet.display['shape'] = 'None'

zero = optics.FlatDetector(orientation=xyz2zxy[:3, :3], zoom=[1, 15, 15], pixsize=0.008)
# zero-order detector based on AXIS, need to adjust for Athena
zero.loc_coos_name = ['zero_x', 'zero_y']
zero.detpix_name = ['zero_x', 'zero_y']

class PerfectAthecat(simulator.Sequence):
    '''Default Definition of Athena with CAT gratings without any misalignments'''

    def add_detectors(self, conf):
        '''Add detectors to the element list

        This is a separate function that is called from __init__ because all
        detectors need different parameters. Placing this specific code in its own
        function makes it easy to override for derived classes.
        '''
        proj2 = analysis.ProjectOntoPlane(orientation=xyz2zxy[:3, :3])
        proj2.loc_coos_name = ['projcirc_y', 'projcirc_z']
        return [RowlandDetArray(conf),
                analysis.ProjectOntoPlane(orientation=xyz2zxy[:3, :3]),
                simulator.Propagator(distance=-1000),
                detcirc,
                proj2,
                simulator.Propagator(distance=-1000),
                flatdet,
                simulator.Propagator(distance=-1000),
                zero]

    def post_process(self):
        self.KeepPos = simulator.KeepCol('pos')
        return [self.KeepPos]

    def __init__(self, conf=conf, **kwargs):
        elem = [spo.aperture,
                spo.SimpleSPOs(conf)]
        if 'gas_kwargs' in conf:
            elem.append(GAS(conf))
        elem.append(FiltersAndQE())
        elem.extend(self.add_detectors(conf))
        elem.append(tagversion)

        super().__init__(elements=elem,
                         postprocess_steps=self.post_process(),
                         **kwargs)
        if ('chirp_energy' in conf) and ('chirp_order' in conf):
            gratings = self.elements_of_class(optics.CATGrating, subclass_ok=False)
            opt = design.bendgratings.NumericalChirpFinder(detcirc,
                                       order=conf['chirp_order'],
                                       energy=conf['chirp_energy'],
                                       d=conf['gas_kwargs']['elem_args']['d'])
            design.bendgratings.chirp_gratings(gratings, opt, conf['gas_kwargs']['elem_args']['d'])


class AthecatForPlot(PerfectAthecat):
    def add_detectors(self, conf):
        return [RowlandDetArray(conf),
                zero]



# Licensed under GPL version 3 - see LICENSE.rst
import copy
import numpy as np
import pytest

from ...optics import FlatDetector, RectangleAperture
from ...design import RowlandTorus
from ..x3d import plot_rays, plot_object

def compare_trimmed(text1, text2):
    for line1, line2 in zip(text1.split('\n'), text2.split('\n')):
        assert line1.strip() == line2.strip()

def test_skip_plot():
    '''Test that plot is skipped with no warning if explicitly
    requested.'''
    det = FlatDetector(zoom=2, position=[2., 3., 4.])
    # First one does not exist, but second one should lead to skip
    # with no warning
    det.display = copy.copy(det.display)
    det.display['shape'] = 'boxxx; None'
    out = plot_object(det)
    out_expected = '''<Scene/>'''
    # check that nothing was added to the scene (i.e. we get an empty scene)
    compare_trimmed(out.XML(), out_expected)

def test_box():
    '''Any box shaped element has the same representation so it's
    OK to test just one of them.'''
    det = FlatDetector(zoom=2, position=[2., 3., 4.])
    det.display = det.display
    det.display['opacity'] = 0.1
    out = plot_object(det)
    out_expected = '''<Scene>
  <Shape>
    <Appearance>
      <Material diffuseColor='1.0 1.0 0.0' transparency='0.9'/>
    </Appearance>
    <IndexedFaceSet colorPerVertex='false' coordIndex='0 2 3 1 -1 4 6 7 5 -1 0 4 6 2 -1 1 5 7 3 -1 0 4 5 1 -1 2 6 7 3 -1' solid='false'>
      <Coordinate point='0.0 1.0 2.0 0.0 5.0 2.0 0.0 1.0 6.0 0.0 5.0 6.0 2.0 1.0 2.0 2.0 5.0 2.0 2.0 1.0 6.0 2.0 5.0 6.0'/>
    </IndexedFaceSet>
  </Shape>
</Scene>
'''
    compare_trimmed(out.XML(), out_expected)

def test_aperture():
    '''Test a plane_with_hole representation.'''
    det = RectangleAperture(zoom=5)
    det.display = det.display
    det.display['opacity'] = 0.3
    out = plot_object(det)
    out_expected = '''<Scene>
  <Shape>
    <Appearance>
      <Material diffuseColor='0.0 0.75 0.75' transparency='0.7'/>
    </Appearance>
    <IndexedTriangleSet colorPerVertex='false' index='0 4 5 0 1 5 1 5 6 1 2 6 2 6 7 2 3 7 3 7 4 3 0 4' solid='false'>
      <Coordinate point='0.0 15.0 15.0 0.0 -15.0 15.0 0.0 -15.0 -15.0 0.0 15.0 -15.0 0.0 5.0 5.0 0.0 -5.0 5.0 0.0 -5.0 -5.0 0.0 5.0 -5.0'/>
    </IndexedTriangleSet>
  </Shape>
</Scene>
'''
    compare_trimmed(out.XML(), out_expected)


def test_surface():
    '''Test an object that's represented as a surface.'''
    rowland = RowlandTorus(R=1000, r=100)
    rowland.display['coo1'] = np.linspace(1, 2, 4)
    rowland.display['coo2'] = np.linspace(-.2, .2, 3)
    out = plot_object(rowland)
    out_expected = '''<Scene>
  <Shape>
    <Appearance>
      <Material diffuseColor='1.0 0.3 0.3' transparency='0.8'/>
    </Appearance>
    <IndexedFaceSet colorPerVertex='false' coordIndex='0 1 4 3 -1 1 2 5 4 -1 3 4 7 6 -1 4 5 8 7 -1 6 7 10 9 -1 7 8 11 10 -1' solid='false'>
      <Coordinate point='1033.0 84.1 -209.4 1054.0 84.1 0.0 1033.0 84.1 209.4 1003.1 97.2 -203.3 1023.5 97.2 0.0 1003.1 97.2 203.3 970.7 99.5 -196.8 990.4 99.5 0.0 970.7 99.5 196.8 939.3 90.9 -190.4 958.4 90.9 0.0 939.3 90.9 190.4'/>
    </IndexedFaceSet>
  </Shape>
</Scene>
'''
    compare_trimmed(out.XML(), out_expected)


def test_rays():
    '''Just two rays to make sure the format is right.'''
    rays = plot_rays(np.arange(12).reshape(2,2,3))
    out_expected = '''<Scene>
  <Shape>
    <Appearance>
      <Material emissiveColor='0.27 0.0 0.33'/>
    </Appearance>
    <LineSet vertexCount='2 2'>
      <Coordinate point='0 1 2 3 4 5 6 7 8 9 10 11'/>
    </LineSet>
  </Shape>
</Scene>
'''
    compare_trimmed(rays.XML(), out_expected)


def test_zip_file(tmp_path):
    """Check that a zipfile is created and contains the right files."""
    zipfile = pytest.importorskip("zipfile", reason="Test writes zip file")
    rays = plot_rays(np.arange(12).reshape(2, 2, 3))
    rays.write_html_archive(base_name=tmp_path / "figure", format="zip")
    assert (tmp_path / "figure.zip").exists()

    zip = zipfile.ZipFile(tmp_path / "figure.zip")
    assert "figure.html" in zip.namelist()
    assert "x3dom.js" in zip.namelist()
    assert "x3dom.css" in zip.namelist()



# Licensed under GPL version 3 - see LICENSE.rst
'''
Try to be lenient in reading, i.e. don't check mandatory keywords.
Will read files that have the correct columns, even if HDUVER or similar
keywords are missing. Need to be stricter?
'''
from warnings import warn
import numpy as np
from astropy.table import QTable, Column, Table
from astropy.io import fits
import astropy.units as u

from marxs.base import TagVersion


def filename_from_meta(filetype='fits', **kwargs):
    '''Generate default filename from ARF or RMF header values

    Parameters
    ----------
    filetype : string
        file ending (the file is usually saved as fits, but the ending
        might be "arf", "rmf", or any other string)
    kwargs: string/int
        Keywords used to construct filename
    '''
    #  convert to string. Will happen anyway when writing to fits
    for k in ['ORDER', 'CCDORDER', 'TRUEORD']:
        #  np.integer is not a subclass of int, so need to test both
        if (k in kwargs) and isinstance(kwargs[k], (int, np.integer)):
            kwargs[k] = f'{kwargs[k]:+}'

    if 'CCDORDER' in kwargs:
        filename = f'_ccdord_{kwargs["CCDORDER"]}_true_{kwargs["TRUEORD"]}'
    else:
        filename = f'_{kwargs["ORDER"]}'

    filename += f'.{filetype}'
    return filename


class OGIPFormatError(Exception):
    pass


class ColOrKeyColumn(Column):
    def __setitem__(self, index, value):
        super().__setitem__(index, value)
        if np.all(self.data == self.data[0]):
            self._parent_table().meta[self.name] = self.data[0]
            self._parent_table().remove_column(self.name)
        else:
            if self.name not in self._parent_table().colnames:
                self._parent_table().columns[self.name] = self
                if self.name in self._parent_table().meta:
                    del self._parent_table().meta[self.name]


class ColOrKeyTable(QTable):
    Column = ColOrKeyColumn

    def __getitem__(self, item):
        if isinstance(item, str) and item in self.meta:
            if item in self.colnames:
                warn(f'{item} is both in meta dict and a column.')
            return self.meta[item]
        else:
            return super().__getitem__(item)

    def __setitem__(self, item, value):
        if isinstance(item, str) and (np.ndim(value) == 0 or
                                      np.all(np.asanyarray(value) == value[0])):
            if np.ndim(value) == 0:
                self.meta[item] = value
            else:
                self.meta[item] = value[0]
            if item in self.colnames:
                self.remove_column(item)
        else:
            super().__setitem__(item, value)


def _check_mandatory_keywords(tab, extra_keywords=[]):
    keywords = ['TELESCOP', 'INSTRUME', 'FILTER'] + extra_keywords
    ind = np.isin(keywords, list(tab.meta))
    if not np.all(ind):
        raise OGIPFormatError('Mandatory keywords missing from meta: {}'.format(np.array(keywords)[~ind]))


def _check_col_and_type(tab, req):
    for col, dtype, ndim in req:
        if not ((col in tab.colnames) or (col in tab.meta)):
            raise OGIPFormatError(f'Required column {col} missing.')
        cancast = [np.can_cast(tab[col].dtype, d, 'unsafe') for d in dtype]
        if not any(cancast):
            raise OGIPFormatError(f'dtype for {col} must be {dtype}.')
        if ((ndim is None) or
            ((ndim == 1) and (col in tab.meta)) or
            (tab[col].ndim == ndim)):
            pass
        else:
            raise OGIPFormatError(f'column {col} must have {ndim} dimensions.')


class ARF(ColOrKeyTable):
    required_cols = [('ENERG_LO', (np.float64,), 1),
                     ('ENERG_HI', (np.float64,), 1),
                     ('SPECRESP', (np.float64,), 1)]

    def write(self, *args, **kwargs):
        self.meta.update({'EXTNAME': 'SPECRESP',
                          'HDUCLASS': 'OGIP',
                          'HDUVERS':  '1.1.0',
                          'HDUCLAS1': 'RESPONSE',
                          'HDUCLAS2': 'SPECRESP',
                          })
        _check_mandatory_keywords(self)
        super().write(*args, **kwargs, format='fits')

    @classmethod
    def read(cls, *args, **kwargs):
        self = super().read(*args, format='fits', **kwargs)
        _check_col_and_type(self, self.required_cols)
        return self


class PHAI(ARF):
    '''


    Parameters
    ----------
    CHANTYPE : 'PHA', 'PI' or None
        If not None, set the CHANTYPE keyword to PHA (should be used if
        possible) or PI. `None` indicates that the CHANTYPE is set already.
    HDUCLAS2 : 'TOTAL', 'NET', 'BKG' or None
        Set HDUCLAS2 to the value given. If none, do not set.



    '''
    # Note: Some dimensions are set to None, which means that checking
    # must be done separately.
    # If dtype is an object (a list for a variable length array), we don't
    # check what's inside that object.
    required_cols = [('CHANNEL', (np.int32,), 1),
                     ('ENERG_HI', (np.float64,), 1),
                            ('N_GRP', (np.int32,), 1),
                            ('F_CHAN', (np.integer, object), None),
                            ('N_CHAN', (np.integer, object), None),
                            ('MATRIX', (np.float64, object), None)]


    def __init__(self, pha, CHANTYPE=None, HDUCLAS2=None):

        super().__init__(self, pha)
        #self.meta['DETCHANS'] = len(self)
        if CHANTYPE is not None:
            self.meta['CHANTYPE'] = CHANTYPE
        if HDUCLAS2 is not None:
            self.meta['HDUCLAS2'] = HDUCLAS2

    def write(self, *args, **kwargs):
        self.meta.update({'EXTNAME': 'SPECTRUM',
                          'HDUCLASS': 'OGIP',
                          'HDUVERS':  '1.2.1',
                          'HDUCLAS1': 'SPECTRUM',
                          'HDUCLAS2': 'SPECRESP',
                          'HDUCLAS3': 'COUNT' if 'COUNTS' in self.colnames else 'RATE',
                          'HDUCLAS4': 'TYPE:I',
                          })
        _check_mandatory_keywords(self,['EXPOSURE', 'POISSERR'])
        for n in ['BACKSCAL', 'AREASCAL', 'QUALITY', 'GROUPING', 'SYS_ERR']:
            if n not in self.colnames and n not in self.meta:
                raise OGIPFormatError('{} must be given as either keyword or column')
        super().write(*args, **kwargs, format='fits')

    @classmethod
    def read(cls, *args, **kwargs):
        self = super().read(*args, format='fits', **kwargs)
        _check_col_and_type(self, self.required_cols)
        return self


class RMF:
    '''Read and represent OGIP RMF data

    Sherpa has a similar class (in fact with more properties)
    I just want to make sure I understand everything that goes into it,
    so I make that myself here.

    Parameters
    ----------
    CHANTYPE : 'PHA', 'PI' or None
        If not None, set the CHANTYPE keyword to PHA (should be used if
        possible) or PI. `None` indicates that the CHANTYPE is set already.
    HDUCLAS3 : 'REDIST', 'DETECTOR', 'FULL' or None
        Set HDUCLAS3 to the value given. If none, do not set.
        Allowed values for HDUCLAS3 are:
           - 'REDIST' for a matrix whose elements represent probabilities
             associated with the photon redistribution process only
           - 'DETECTOR' for a matrix whose elements have been multipled by
             all energy-dependent effects associated with detector
             (eg detector efficiency, window transmission etc).
           - 'FULL' for a matrix whose elements have been multipled by all
             energy-dependent effects associated with detector, optics,
             collimator, filters etc.

    Parameters
    ----------
    rmffile : filename
    '''
    # Note: Some dimensions are set to None, which means that checking
    # must be done separately.
    # If dtype is an object (a list for a variable length array), we don't
    # check what's inside that object.
    matrix_required_cols = [('ENERG_LO', (np.float64,), 1),
                            ('ENERG_HI', (np.float64,), 1),
                            ('N_GRP', (np.int32,), 1),
                            ('F_CHAN', (np.int32, object), None),
                            ('N_CHAN', (np.int32, object), None),
                            ('MATRIX', (np.float64, object), None)]
    ebounds_required_cols = [('CHANNEL', (np.int32,), 1),
                             ('E_MIN', (np.float64,), 1),
                             ('E_MAX', (np.float64,), 1)]

    def __init__(self, matrix, ebounds, CHANTYPE=None, HDUCLAS3=None):
        self.matrix = matrix
        self.ebounds = ebounds
        for m in [self.matrix, self.ebounds]:
            m.meta['DETCHANS'] = len(m)
            if CHANTYPE is not None:
                m.meta['CHANTYPE'] = CHANTYPE
            if HDUCLAS3 is not None:
                m.meta['HDUCLAS3'] = HDUCLAS3
        _check_col_and_type(self.matrix, self.matrix_required_cols)
        _check_col_and_type(self.ebounds, self.ebounds_required_cols)
        _check_mandatory_keywords(self.matrix, extra_keywords=['CHANTYPE'])
        _check_mandatory_keywords(self.ebounds, extra_keywords=['CHANTYPE'])

    @classmethod
    def read(cls, rmffile):
        matrix = ColOrKeyTable.read(rmffile, format='fits', hdu='MATRIX')
        ebounds = ColOrKeyTable.read(rmffile, format='fits', hdu='EBOUNDS')
        _check_col_and_type(matrix, cls.matrix_required_cols)
        _check_col_and_type(ebounds, cls.ebounds_required_cols)
        for tab in [matrix, ebounds]:
            _check_mandatory_keywords(tab, ['CHANTYPE', 'DETCHANS'])
        return cls(matrix, ebounds)

    @staticmethod
    def ebounds_from_edges(channel_edges):
        '''
        Parameters
        ----------
        waveedges : `~astropy.units.quantity.Quantity`
            Edges of channels
        '''
        energy_edges = channel_edges.to(u.keV,
                                        equivalencies=u.spectral()).value
        sort_energies = np.argsort(energy_edges)
        ebounds = ColOrKeyTable({'CHANNEL': np.arange(len(channel_edges) - 1,
                                                      dtype=np.int32),
                                 'E_MIN': energy_edges[sort_energies][:-1],
                                 'E_MAX': energy_edges[sort_energies][1:]})
        for col in ['E_MIN', 'E_MAX']:
           ebounds[col].unit = u.keV
        return ebounds

    def write(self, rmffile, overwrite=False, TLMIN=None):
        '''
        Paramters
        ---------
        TLMIN : int or `None`
            If `None` then the TLMIN# keywords should already be set, if not,
            if is a convenient way to set them on writing.
       '''
        _check_col_and_type(self.matrix, self.matrix_required_cols)
        _check_col_and_type(self.ebounds, self.ebounds_required_cols)
        matrix_tlmin = 'TLMIN{}'.format(self.matrix.colnames.index('F_CHAN') + 1)
        eb_tlmin = 'TLMIN{}'.format(self.ebounds.colnames.index('CHANNEL') + 1)
        if TLMIN is not None:
            self.matrix.meta[matrix_tlmin] = TLMIN
            self.ebounds.meta[eb_tlmin] = TLMIN
        _check_mandatory_keywords(self.matrix, ['CHANTYPE', 'DETCHANS',
                                                matrix_tlmin])
        _check_mandatory_keywords(self.ebounds, ['CHANTYPE', 'DETCHANS',
                                                 eb_tlmin])
        self.matrix.meta['EXTNAME'] = 'MATRIX'
        self.matrix.meta['HDUCLASS'] = 'OGIP'
        self.matrix.meta['HDUVERS'] = '1.1.0'
        self.matrix.meta['HDUCLAS1'] = 'RESPONSE'
        self.matrix.meta['HDUCLAS2'] = 'RSP_MATRIX'
        # further required keywords: CHANTYPE (PI or PHA), DETCHANS, TLMIN# (# is column number of F_CHAN)
        # Recommended keywords: NUMGRP, NUMELT


        # EBOUNDS
        self.ebounds.meta['EXTNAME'] = 'EBOUNDS'
        self.ebounds.meta['HDUCLASS'] = 'OGIP'
        self.ebounds.meta['HDUCLAS1'] = 'RESPONSE'
        self.ebounds.meta['HDUCLAS2'] = 'EBOUNDS'
        self.ebounds.meta['HDUVERS'] = '1.2.0'

        # Here I'm assuming that if MATRIX is an object (a list) then
        # F_CHAN and N_CHAN are also. That's the case when this class
        # is used ot make them, but might not be true in general if
        # read in from a file.
        if self.matrix['MATRIX'].dtype == object:
            self.variable_length_to_fixed_length()

        hdulist = fits.HDUList([fits.PrimaryHDU(),
                                fits.table_to_hdu(self.matrix),
                                fits.table_to_hdu(self.ebounds)])

        hdulist.writeto(rmffile, overwrite=overwrite, checksum=True)

    def row(self, energy):
        with u.set_enabled_equivalencies(u.spectral()):
            rowind = (energy >= self.matrix['ENERG_LO']) & \
                (energy < self.matrix['ENERG_HI'])
        rows = rowind.nonzero()[0][0]
        return self.matrix[rows]

    @property
    def en_mid(self):
        return 0.5 * (self.ebounds['E_MIN'] + self.ebounds['E_MAX'])

    def full_rmf(self, energy):
        '''Return rmf over the full range of channels

        Parameters
        ----------
        energy : `astropy.quantity.Quantity`

        Returns
        -------
        e_mid : array
            mid-points of energy bins in keV
        rmf : array
            rmf value in each bin
        '''
        rmf = np.zeros(len(self.en_mid))
        row = self.row(energy)
        tlmin = self.matrix.meta['TLMIN{}'.format(self.matrix.colnames.index('F_CHAN') + 1)]
        for i in range(row['N_GRP']):
            # Python is 0 indexed, FITS is 1 indexed
            chans = slice(row['F_CHAN'][i] - tlmin,
                          row['F_CHAN'][i] + row['N_CHAN'][i] - tlmin)
            matindex = np.cumsum(np.concatenate(([0], row['N_CHAN'])))
            mat = slice(matindex[i], matindex[i + 1])
            rmf[chans] = row['MATRIX'][mat]
        return self.ebounds['E_MIN'], self.ebounds['E_MAX'], rmf

    def rmf(self, energy):
        '''Return rmf over channels where values is non-zero

        Parameters
        ----------
        energy : `astropy.quantity.Quantity`

        Returns
        -------
        e_min, e_max : array
            min and max of energy bins in keV
        rmf : array
            rmf value in each bin
        '''
        row = self.row(energy)
        ind = np.concatenate([np.arange(row['F_CHAN'][i] - 1,
                                        row['F_CHAN'][i] + row['N_CHAN'][i]-1)
                              for i in range(row['N_GRP'])])
        return self.ebounds['E_MIN'][ind], self.ebounds['E_MAX'][ind], \
            np.asanyarray(row['MATRIX'])

    def rmf_angstrom(self, energy):
        en_lo, en_hi, rmf = self.rmf(energy)
        return en_hi.to(u.Angstrom, equivalencies=u.spectral()), en_lo.to(u.Angstrom, equivalencies=u.spectral()), rmf

    @staticmethod
    def arr_to_rmf_matrix_row(arr, TLMIN_F_CHAN, threshold=1e-6):
        '''Split an array into channel groups for an RMF

        The format of an RMF is set by the [OGIP cal/gen/92-002]_.
        For efficiency, not the entire RMF matrix is stored in the
        rmf files, but only the non-zero components. So, for each
        row (each row contains the matrix for one energy channel)
        the elements that are zero or below a threshold are removed
        from the matrix. This leaves one or more "channel subsets"
        and extra index arrays are needed to specify the position
        and lengths of these channel subsets in the full array.

        [OGIP cal/gen/92-002]_ discusses how arrays can be stored
        as fixed-length or as variable-length arrays in a fits file.
        Since this function operates on a single row only, it return
        lists. Formatting in a fixed-length array, if needed, has
        to be done separately.

        References
        ----------
        .. [OGIP cal/gen/92-002] https://heasarc.gsfc.nasa.gov/docs/heasarc/caldb/docs/memos/cal_gen_92_002/cal_gen_92_002.html

        Parameters
        ----------
        arr : `np.array`
            RMF respsonse matrix over all channels of the isntrument
        threshold : float
            Array values below the threshold will be removed.

        Returns
        -------
        N_GRP : int
            Number of channel subsets
        F_CHAN : list
            First channel in each subset (1 indexed)
        N_CHAN : list
            Number of channels in each subset
        MATRIX : list
            Matrix elements
        '''
        ind = arr > threshold
        # Add False at beginning to end, so that np.diff can be
        # used to detect changes to and from True even if the first
        # and / or last value in the row is True
        nfchan = np.diff(np.concatenate(([False], ind, [False]))).nonzero()[0]
        start_end = nfchan.reshape((-1, 2))
        nchan = start_end[:, 1] - start_end[:, 0]
        #
        fchan = start_end[:, 0] + TLMIN_F_CHAN
        assert nchan.shape == fchan.shape
        return len(fchan), list(fchan), list(nchan), list(arr[ind])

    def variable_length_to_fixed_length(self):
        '''Convert variable length arrays to fixed length arrays

        Astropy cannot write variable length arrays
        (https://github.com/astropy/astropy/issues/11323 and linked isues).
        Until that is fixed, all arrays need to be converted to fixed length
        for writing.

        Note: In many cases, this will only moderately increase the file size
        but speed up operations significantly, so it might be a good idea
        anyway.
        '''
        m = self.matrix
        # int16 is recommended in OGIP, but too small for e.g. Arcus
        # So, determine smallest inttype that works
        maxvalf = np.concatenate(m['F_CHAN']).max()
        maxvaln = np.concatenate(m['N_CHAN']).max()
        maxval = max(maxvalf, maxvaln)
        for inttype in [np.int16, np.int32, np.int64]:
            if inttype(maxval) == maxval:
                break
        fchan = np.zeros((len(m), np.max(m['N_GRP'])), dtype=inttype)
        nchan = np.zeros_like(fchan)
        matrix = np.zeros((len(m), max([sum(r) for r in m['N_CHAN']])),
                          dtype=np.float32)
        for i, r in enumerate(m):
            fchan[i, :len(r['F_CHAN'])] = r['F_CHAN']
            nchan[i, :len(r['N_CHAN'])] = r['N_CHAN']
            matrix[i, :len(r['MATRIX'])] = r['MATRIX']
        # if max(N_GRP) == 1 then we can squueze out the first dimension
        # and write a number per row instead of an array.
        m['F_CHAN'] = fchan.squeeze()
        m['N_CHAN'] = nchan.squeeze()
        m['MATRIX'] = matrix

    @staticmethod
    def from_Gauss_sigma(bin_edges, sigmatab, threshold=1e-6,
                         headerkw={'SATELLIT': 'unknown', 'TELESCOP': 'unknown',
                                   'INSTRUME': 'unknown', 'FILTER': 'unknown'}):
        '''Make RMF for a Gaussian redistribution

        This is a convenience function for a common, simple case.
        For more complex parameterization of the redistributions, see
        `marxs.reduction.lsfparm`.

        Parameters
        ----------
        bin_edges : `~astropy.units.quantity.Quantity`
            Edges of the channels in energy or wavelength. The RMF that is
            generated from n edges will have n-1 bins.
        sigmatab : `~astropy.units.table.Table`
            sigma for a Gaussian in a table with columns ``'wave'``
            and ``'sigma_wave'`` or ``'energy'`` and ``sigma_energy'``.
            In the latter case, the information is converted to wavelength space.
            Note that a Gaussian in energy space is not the same as in wavelength space,
            so this should only be used if this approximate treatment is sufficient.
        threshold : float
            To reduce the file size, RMF components below the threshold value
            are cut-off
        headerkw : dict
            Keywords arguments for the header of the RMF file. Minimum required list
            are the header keywords SATELLIT, TELESCOP, INSTRUME, FILTER
        '''
        # This module requires sherpa, so we import it down here.
        # That way, the rest of this class can be used even when Sherpa is not installed.
        from marxs.reduction.lsfparm import make_rmf
        if 'energy' in sigmatab.colnames and 'wave' not in sigmatab.colnames:
            sigmatab['wave'] = sigmatab['energy'].to(u.Angstrom, equivalenicies=u.spectral())
            sigmatab['sigma_wave'] = sigmatab['wave'] * sigmatab['sigma_energy'] / sigmatab['energy']
        sigmatab.sort('wave')

        lsfparm = Table({'NUM_WIDTHS': [1],
                         # We are not using extraction width here, but the format requires a number
                         'WIDTHS': [[1.]],
                         'NUM_LAMBDAS': [len(sigmatab)],
                         'EE_FRACS': [np.ones((1, len(sigmatab)))],
                         'LAMBDAS': [sigmatab['wave']],
                         'GAUSS1_PARMS': [np.stack([np.ones(len(sigmatab)),
                                           sigmatab['sigma_wave'].to(u.Angstrom).value,
                                           sigmatab['wave'].to(u.Angstrom).value
                                           ]).T[None, :, :]]},
                         units=[None, u.deg, None, None,  u.Angstrom, None]
                        )
        tagversion = TagVersion(**headerkw)
        tagversion(lsfparm)
        return make_rmf(lsfparm[0],
                        bin_edges.to(u.Angstrom, equivalencies=u.spectral()),
                        1 * u.deg,
                        threshold=threshold,
                        kw_interp={})


import numpy as np

from marxs.missions import chandra
from marxs.source import ConstantPointSource, FixedPointing
from marxs.optics import MarxMirror
from marxs.analysis import find_best_detector_position, measure_FWHM

mysource = ConstantPointSource((30., 30.), energy=1., flux=50.)
mypointing = FixedPointing(coords=(30, 30.))
# mypointing = chandra.LissajousDither(coords=(30.,30.), roll=0.)
marxm = MarxMirror('./marxs/optics/hrma.par', position=np.array([0., 0,0]))
hetg = chandra.HETG()
acis = chandra.ACIS(chips=[4,5,6,7,8,9], aimpoint=chandra.AIMPOINTS['ACIS-S'])
#mydet = FlatDetector(zoom=1e5, pixsize=23.985e-3)

# Chandra class is not finished yet.
Chandra = chandra.Chandra(sequence=[mypointing, marxm, hetg, acis])

photons = mysource.generate_photons(5000)
#photons = Chandra(photons)

photons = mypointing(photons)
photons = marxm(photons)
photons = photons[photons['probability'] > 0]

photons = hetg(photons)
photons['hetgy'] = photons['pos'].data[:,1]/photons['pos'].data[:,3]
photons['hetgz'] = photons['pos'].data[:,2]/photons['pos'].data[:,3]

photons = acis(photons)
mypointing.write_asol(photons, 'asol.fits')
Chandra.write_evt(photons, 'photons.fits')

p = photons[(photons['order']==-1) & (photons['facet'] > 194)]
xbest = find_best_detector_position(p, col='det_x')
print(xbest.x)
print(measure_FWHM(p['det_x']), measure_FWHM(p['det_y']))

#for x in [-50, 0, 50, xbest.x]:
#    mdet = FlatDetector(position=np.array([x, 0, 0]), zoom=1e5, pixsize=1.)
#    photons = mdet.process_photons(photons)
#    photons.rename_column('det_x', 'det_x_{0}'.format(x))
#    photons.rename_column('det_y', 'det_y_{0}'.format(x))


import astropy
# Make output for Glue
# output non-vector columns
pout = photons[:]
for c in pout.colnames:
    if len(pout[c].shape) > 1:
        pout.remove_column(c)

# output facet information
facet = np.arange(len(hetg.elements))
facet_tab = astropy.table.Table({'facet':facet, 'facet_x': facet, 'facet_y': facet, 'facet_z':facet})
for i in facet:
    facet_tab['facet_x'][i] = hetg.elements[i].pos4d[0, 3]
    facet_tab['facet_y'][i] = hetg.elements[i].pos4d[1, 3]
    facet_tab['facet_z'][i] = hetg.elements[i].pos4d[2, 3]

photfac = astropy.table.join(pout, facet_tab)
photfac.write('chandra.fits', overwrite=True)


mfits = astropy.table.Table.read('point.fits')
for c in mfits.colnames:
    if len(mfits[c].shape) > 1:
        mfits.remove_column(c)

mfits.write('marx.fits', overwrite=True)



import os
import configparser
import subprocess
import logging

from marxs.base import TagVersion

__all__ = ['config', 'tabversion', 'id_num_offset',
           'git_hash', 'git_info',
           ]

logger = logging.getLogger(__name__)

id_num_offset = {'1': 0,
                 '2': 1000,
                 '1m': 10000,
                 '2m': 11000}

config = configparser.ConfigParser()
confs_found = config.read(['arcus.cfg',
                           os.path.expanduser('~/.astropy/config/arcus.cfg')
                           ])

# subprocess uses fork internally, which makes a child process
# that essentially makes a copy of everything that python has
# in memory, essentially doubling the memory footprint of python
# at that point. For long running scripts with big simulations that
# can be enough to run out of memory.
# So, just get this info once during the first import and save in
# module-level variables.
git_hash = b'no arcus cal data found'
git_info = b'no arcus cal data found'
if len(confs_found) > 0:
    logger.info(f'Found arcus.cfg in {confs_found}')
    try:
        gh = subprocess.check_output(["git", "describe", "--always"],
                                      cwd=config['data']['caldb_inputdata'])[:-1]

        gi = subprocess.check_output(['git', 'show', '-s', '--format=%ci',
                                      gh],
                                      cwd=config['data']['caldb_inputdata'])

        git_hash = gh.decode()[:10]
        git_info = gi.decode()[:19]
    except Exception as err:
        logger.warning(f"Error trying to read {config['data']['caldb_inputdata']}: {err}")


tagversion = TagVersion(SATELLIT=('ARCUS', 'placeholder - no name registered with OGIP'),
                        TELESCOP=('ARCUS', 'placeholder - no name registered with OGIP'),
                        INSTRUME=('ARCUSCAM',  'placeholder - no name registered with OGIP'),
                        FILTER=('NONE', 'filter information'),
                        GRATING=('ARCUSCAT',  'placeholder - no name registered with OGIP'),
                        ARCDATHA=(git_hash, 'Git hash of simulation input data'),
                        ARCDATDA=(git_info, 'Commit time of simulation input data'),
)



# Licensed under GPL version 3 - see LICENSE.rst
import os
import logging

from astropy.table import Table
import astropy.units as u
from .utils import config, git_hash, git_info

hash_displayed = False


class DataFileFormatException(Exception):
    pass


def log_tab_metadata(dirname, filename):
    '''Print information about loaded files to standard out.

    Parameters
    ----------
    dirname : string
        Name for the directory in the caldb-input file structure
    filename : string
        Name of data file (without the ".csv" part)
    '''
    global hash_displayed
    if not hash_displayed:
        text = 'data files in {}: git hash {} (committed on {})'
        logging.info(text.format(config['data']['caldb_inputdata'],
                                 git_hash, git_info))
        hash_displayed = True
    logging.info('Loading data from {0}/{1}'.format(dirname, filename))


def load_number(dirname, filename, valuename):
    '''Get a single number from an ecsv input file

    Parameters
    ----------
    dirname : string
        Name for the directory in the caldb-input file structure
    filename : string
        Name of data file (without the ".csv" part)
    valuename : string
        Name of the column that hold the value

    Returns
    -------
    val : float or `astropy.units.Quantity`
        If the unit of the column is set, returns a `astropy.units.Quantity`
        instance, otherwise a plain float.
    '''
    tab = Table.read(os.path.join(config['data']['caldb_inputdata'], dirname,
                                  filename + '.csv'), format='ascii.ecsv')
    log_tab_metadata(dirname, filename)
    if len(tab) != 1:
        raise DataFileFormatException('Table {} contains more than one row of data.'.format(filename))
    else:
        if tab[valuename].unit is None:
            return tab[valuename][0]
        else:
            return u.Quantity(tab[valuename])[0]


def load_table(dirname, filename):
    '''Get a table from an ecsv input file

    Parameters
    ----------
    dirname : string
        Name for the directory in the caldb-input file structure
    filename : string
        Name of data file (without the ".csv" part)

    Returns
    -------
    val : `astropy.table.Table`
    '''
    tab = Table.read(os.path.join(config['data']['caldb_inputdata'], dirname,
                                  filename + '.csv'), format='ascii.ecsv')
    log_tab_metadata(dirname, filename)
    return (tab)



# Licensed under GPL version 3 - see LICENSE.rst
'''This module contains functions to modify gratings.

After flat simple gratings have been placed on a Rowland-torus, they can be
changed to more complicated geometries or set-ups. For example, placing just
the center of the grating on the Rowland-torus, can leave the edges to deviate.
In this case, gratings might be bend in one dimension or a more complex
arrangement of grating bars (a "chirp") could be specified. Conceptually, it
is often useful to think of this as a multi-step process for the simulation,
so in practice a chirp would have to be known in grating manufacturing.

To support this conceptual idea (place the gratings, then rotate it, then do X,
then chirp it) the functions in this module operate on exciting grating objects.
'''
import numpy as np
from scipy.optimize import minimize_scalar
from scipy.interpolate import RectBivariateSpline
from transforms3d.affines import decompose, decompose44

from marxs.math.geometry import Cylinder
from marxs.math.utils import h2e, e2h, norm_vector
from marxs.utils import generate_test_photons
from marxs.optics import OrderSelector


def bend_gratings(gratings, radius):
    '''Bend gratings to follow the Rowland cirle.

    Gratings are bend in one direction (the dispersion direction) only.

    The numerical procedure used to calculate the bending takes the central ray as
    a fixed ppoint assuming that the central ray always goes to the correct position!

    Parameters
    ----------
    gratings : list
        List of gratings to be bend
    radius : float
        Radius of the newly bend gratings
    '''
    for e in gratings:
        t, rot, z, s = decompose(e.geometry.pos4d)
        d_phi = np.arctan(z[1] / radius)
        c = Cylinder({'position': t - radius * h2e(e.geometry['e_x']),
                      'orientation': rot,
                      'zoom': [radius, radius, z[2]],
                      'phi_lim': [-d_phi, d_phi]})
        c._geometry = e.geometry._geometry
        e.geometry = c
        e.display['shape'] = 'surface'
        for e1 in e.elements:
            # can't be the same geometry, because groove_angle is part of _geometry and that's different
            # Maybe need to get that out again and make the geometry strictly the geometry
            # But for now, make a new cylinder of each of them
            # Even now, not sure that's needed, since intersect it run by FlatStack
            c = Cylinder({'position': t - radius * h2e(e.geometry['e_x']),
                          'orientation': rot,
                          'zoom': [radius, radius, z[2]],
                          'phi_lim': [-d_phi, d_phi]})
            c._geometry = e1.geometry._geometry
            e1.geometry = c
            e1.display['shape'] = 'surface'


class NumericalChirpFinder():
    '''Optimizer to determine optimal chirp by ray-tracing individual rays

    This object passes a ray through the center of a grating with the specified energy and
    diffraction order. It records the position of this center ray. Then, rays are passed through the grating at different positions. For each position, the grating period is optimized numerically, such that these test rays hit the same position as the center ray, when the object is called is called.

    The purpose of wrapping this in an object (as opposed to a simple function) is that certain settings such as energy and order of diffraction are set when the oject is initialized



    Assumes that the focal point (=position of the 0th order) is at the
    origin of the coordinate system.

    Parameters
    ----------
    detector : marxs element
        Mamrx
    colname : string
        Name of column in photon list that the detector ``detector`` writes in
    '''
    uv = [0, 0]

    def __init__(self, detector, order, energy, d=0.0002,
                 colname='detcirc_phi'):
        self.photon = generate_test_photons(1)
        self.detector = detector
        self.energy = energy
        self.order = order
        self.base_d = d
        self.colname = colname

    def set_grat(self, grat):
        self.grat = grat
        self.calc_goal()

    def set_uv(self, uv):
        self.uv = uv
        self.init_photon()

    def calc_goal(self):
        if not hasattr(self.grat, 'original_orderselector'):
            self.grat.original_orderselector = self.grat.order_selector
        self.grat.order_selector = OrderSelector([self.order])
        self.grat._d = self.base_d
        self.set_uv([0., 0.])
        self.run_photon()
        self.goal = self.photon[self.colname][0]

    def posongrat(self):
        pos = h2e(self.grat.geometry['center'] +
                  self.uv[0] * self.grat.geometry['v_y'] +
                  self.uv[1] * self.grat.geometry['v_z'])
        return pos

    def init_photon(self):
        pos = self.posongrat()
        self.pos = e2h(1.1 * pos, 1)
        self.dir = norm_vector(- e2h(pos.reshape(1, 3), 0))
        self.reset_photon()

    def reset_photon(self):
        self.photon['pos'] = self.pos
        self.photon['dir'] = self.dir
        self.photon['probability'] = 1

    def run_photon(self):
        self.reset_photon()
        self.photon = self.grat(self.photon)
        self.photon = self.detector(self.photon)

    def optimize_func(self, d):
        self.grat._d = d * self.base_d
        self.run_photon()
        return np.abs(self.photon[self.colname][0] - self.goal)

    def correction_on_d(self, uarray=np.array([-.999, 0, .999]),
                        varray=np.array([0])):
        corr = np.ones((len(uarray), len(varray)))
        for j, u in enumerate(uarray):
            for k, v in enumerate(varray):
                self.set_uv([u, v])
                corr[j, k] = minimize_scalar(self.optimize_func,
                                             bracket=(.99, 1., 1.01)).x
        return corr

    def __call__(self, grat, *args, **kwargs):
        self.set_grat(grat)
        return self.correction_on_d(*args, **kwargs)


class BendNumericalChirpFinder(NumericalChirpFinder):
    def posongrat(self):
        trans, rot, zoom, shear = decompose44(self.grat.geometry.pos4d)
        p_lim = self.grat.geometry.phi_limits
        p_center = np.mean(p_lim)
        # Not accounting for 2 pi wrap and crazy stuff
        p_half = (p_lim[1] - p_lim[0]) / 2
        pos = h2e(self.grat.geometry.parametric_surface([p_center + self.uv[0] * p_half],
                                                        [self.uv[1] * zoom[2]]))
        return pos[0, 0, :]


def chirp_gratings(gratings, optimizer, d,
                   uarray=np.array([-.999, 0, .999]), varray=np.array([0])):
    '''
    Parameters
    ----------
    gratings : list

    optimizer : callable
    '''
    for grat in gratings:
        corr = optimizer(grat, uarray, varray)
        corr = np.tile(corr[:, 0], (3, 1)).T
        ly = np.linalg.norm(grat.geometry['v_y'])
        lz = np.linalg.norm(grat.geometry['v_z'])
        grat.fitted_d_corr = corr
        grat.spline = RectBivariateSpline(ly * uarray, lz * uarray,
                                          d * corr,
                                          bbox=[-ly, ly, -lz, lz],
                                          kx=2, ky=2)

        def func(self, intercoos):
            return self.spline(intercoos[:, 0], intercoos[:, 1], grid=False)

        # Invoking the descriptor protocol to create a bound method
        # see https://stackoverflow.com/questions/972/adding-a-method-to-an-existing-object-instance
        grat._d = func.__get__(grat)
        grat.order_selector = grat.original_orderselector



# Licensed under GPL version 3 - see LICENSE.rst

import numpy as np
import pytest

from ...utils import generate_test_photons
from ..base import _assign_col_value


def test_assign_col():
    'Test a scalar and a vector case'
    photons = generate_test_photons(5)
    # scaler index
    _assign_col_value(photons, 'energy', 2, 3.)
    assert np.allclose(photons['energy'], [1, 1, 3, 1, 1])
    # array index
    _assign_col_value(photons, 'energy', photons['energy'] < 2, 2)
    assert np.allclose(photons['energy'], [2, 2, 3, 2, 2])

    _assign_col_value(photons, 'probability', 1, .5)
    _assign_col_value(photons, 'probability', photons['energy'] < 2.5, .5)
    assert np.allclose(photons['probability'], [.5, .25, 1, .5, .5])


def test_assign_col_probability_wrong():
    'And exception should be thrown for probabilities below 0 or above 1'
    photons = generate_test_photons(5)

    with pytest.raises(ValueError):
        _assign_col_value(photons, 'probability', 1, 1.5)
    with pytest.raises(ValueError):
        _assign_col_value(photons, 'probability', photons['energy'] < 2.5, -.1)



import numpy as np
from astropy.table import Table
import pytest

from ...utils import generate_test_photons
from ...optics import ThinLens, FlatGrating, FlatDetector, OrderSelector
from ..gratings import (resolvingpower_from_photonlist,
                        resolvingpower_from_photonlist_robust,
                        effectivearea_from_photonlist,
                        identify_photon_in_subaperture,
                        average_R_Aeff,
                        CaptureResAeff,
                        CaptureResAeff_CCDgaps
                        )

def test_resolvingpower_from_photonlist():
    '''Regression test.

    [#159] left behind an undefined ``filterfunc``. This is fixed and a
    regression test added [#165].
    '''
    p = generate_test_photons(1000)
    for i in [1, 2]:
        p['pos'][:, i] = -0.1 + 0.2 * np.random.random(len(p))
    lens = ThinLens(focallength=1)
    p = lens(p)
    orders = np.array([-1, 0, 1])
    grat = FlatGrating(position=[-0.01, 0, 0], d=1e-4,
                       order_selector=OrderSelector(orders))
    p = grat(p)
    det = FlatDetector(position=[-1, 0, 0])
    p = det(p)
    res0, pos0, std0 = resolvingpower_from_photonlist(p, orders, col='det_x')
    assert res0[0] > 10
    assert res0[1] < 1
    assert res0[2] > 10
    res, pos, std = resolvingpower_from_photonlist(p, orders, col='det_x', zeropos=0)
    assert res[0] > 10
    assert res[1] < 1
    assert res[2] > 10
    # y is the cross-dispersion direction...
    resy, posy, stdy = resolvingpower_from_photonlist(p, orders, col='det_y', zeropos=0)
    assert np.all(resy < 1.)
    assert np.allclose(posy, 0, atol=1e-4)


def test_resolvingpower_from_photonlist_robust():
    orders = [1, 2, 3]
    size = 10000
    photons = Table({'col1': 2000 + np.random.normal(size=size),
                     'col2': 1000 + np.random.normal(size=size),
                     'order': np.random.choice(orders, size=size)})
    res, pos, std = resolvingpower_from_photonlist_robust([photons, photons],
                                                          orders,
                                                          ['col1', 'col2'],
                                                          [0, 0])
    assert 1000 / 2.3548 * np.ones(3) == pytest.approx(res, rel=1e-1)
    assert np.ones(3) == pytest.approx(std, rel=1e-1)
    assert 1000 * np.ones(3) == pytest.approx(pos, rel=1e-1)

    res, pos, std = resolvingpower_from_photonlist_robust([photons['order', 'col1'],
                                                           photons['order', 'col2']],
                                                           orders,
                                                          ['col1', 'col2'],
                                                          [0, -1000])
    assert 2000 / 2.3548 * np.ones(3) == pytest.approx(res, rel=1e-1)

def test_resolvingpower_from_photonlist_robust2():
    '''Now give zeropos for one of the distributions'''
    size=10000
    photons = Table({'col1': 2000 + np.random.normal(size=size),
                     'col2': 1000 + np.random.normal(size=size),
                     'ord': np.random.choice([0, 1, 2], size=size)})
    photons['col1'][photons['ord'] == 0] = 0
    res, pos, std = resolvingpower_from_photonlist_robust([photons, photons],
                                                          [1,2],
                                                          ['col1', 'col2'],
                                                          [None, 10000],
                                                          ordercol='ord')
    assert 2000 / 2.3548 * np.ones(2) == pytest.approx(res, rel=1e-1)
    assert np.ones(2) == pytest.approx(std, rel=1e-1)
    assert 2000 * np.ones(2) == pytest.approx(pos, rel=1e-1)



def test_aeff_from_photonlist():
    '''Check the effective area calculation'''
    p = generate_test_photons(5)
    p['a'] = [0, 1, 2, 2, 1]
    p['probability'] = [.1, 1., 1., .3, .5]
    # orders in random order, just in case
    aeff = effectivearea_from_photonlist(p, [2, 1, 0, -1], 5,
                                         A_geom=2., ordercol='a')
    assert aeff[3] == 0
    assert np.isclose(aeff[2], .04)
    assert np.isclose(aeff[1], .6)


def test_identify_photon_in_subaperture():
    angle = [0, 59, 61, 180, 289]
    inaper = identify_photon_in_subaperture(np.deg2rad(angle),
                                            np.deg2rad(30), ang_0=np.pi / 2)
    assert np.all(inaper == [False, False, True, False, True])


def test_average_R_Aeff():
    r_out, aeff_out = average_R_Aeff(np.array([np.nan, np.nan]),
                                     np.array([1, 2]))
    assert r_out is np.ma.masked
    assert aeff_out == 3
    r_out, aeff_out = average_R_Aeff(np.array([5., 5.]),
                                     np.array([1, 2]))
    assert r_out == pytest.approx(5)
    r_out, aeff_out = average_R_Aeff(np.array([5., 2.]),
                                     np.array([1, 2]))
    assert r_out == pytest.approx(3)

    nan = np.nan
    r = np.array([
       [[3.72047331e+03, nan, nan, 1.22772588e-02],
        [3.68709325e+03,            nan,            nan, 5.16674211e-02],
        [3.78030863e+03,            nan,            nan, 5.68255010e-02],
        [3.67658767e+03,            nan, 1.20839326e+03, 2.84897083e-02]],

       [[3.63193038e+03,            nan,            nan, 6.43244337e-05],
        [           nan,            nan,            nan, 2.55943724e-02],
        [3.81094245e+03,            nan,            nan, 3.20472833e-02],
        [3.62704572e+03,            nan,            nan, 4.04451453e-02]]])
    aeff = np.array([
       [[45.43284984,  0.        ,  0.        ,  4.14357052],
        [46.01057067,  0.        ,  0.        ,  3.95845344],
        [45.35928564,  0.        ,  0.        ,  4.00130456],
        [44.54923056,  0.        ,  8.66382063,  4.11383244]],

       [[44.93080722,  0.        ,  0.        ,  4.35276434],
        [ 0.        ,  0.        ,  0.        ,  3.96159335],
        [45.1660785 ,  0.        ,  0.        ,  3.68142596],
        [44.43211355,  0.        ,  0.        ,  4.3622082 ]]])
    expected_r = np.ma.masked_invalid(
       [[3716.1898154882915, nan, 1208.3932636420263, 0.03699616674607531],
        [3690.417678741742, nan, nan, 0.024213505777004152]])
    expected_aeff = np.array([
        [181.35193671,   0.        ,   8.66382063,  16.21716095],
        [134.52899927,   0.        ,   0.        ,  16.35799185]])
    r_out, aeff_out =  average_R_Aeff(r, aeff, axis=1)
    assert np.allclose(r_out, expected_r, equal_nan=True)
    assert np.allclose(aeff_out, expected_aeff, equal_nan=True)


def test_capture_res_aeff():
    '''Test the captures res/aeff class.

    Similar to the previous test, this is not a complete functional test,
    but it checks the interfaces.
    The actual function to calculate the effective area is tested elsewhere.
    '''
    p = generate_test_photons(200)
    p['order'] = 0
    p['order'][50:] = 5
    p['xpos'] = -100
    p['xpos'][50:] = np.random.normal(scale=1, size=150)

    resaeff = CaptureResAeff(A_geom=10., order_col='order',
                             orders=[0, 2, 5], dispersion_coord='xpos')
    out = resaeff(p)
    assert np.allclose(out['Aeff'], [2.5, 0., 7.5])
    assert np.isclose(out['Aeff0'], 2.5)
    assert np.isclose(out['Aeffgrat'], 7.5)
    assert len(out['R']) == 3
    assert np.isnan(out['R'][1])
    # Large rtol to reduce risk of random failures
    assert np.isclose(out['R'][2], 50., rtol=.8)


def test_capture_res_aeff_filter():
    '''Ensure that photons with probability 0 will be ignored.
    '''
    p = generate_test_photons(200)
    p['probability'] = 0
    p['order'] = 0
    p['order'][50:] = 5
    p['xpos'] = -100
    p['xpos'][50:] = np.random.normal(scale=1, size=150)

    resaeff = CaptureResAeff(A_geom=10., order_col='order',
                             orders=[0, 2, 5], dispersion_coord='xpos')
    out = resaeff(p)
    assert np.all(out['Aeff'] == 0)
    assert out['Aeff0'] == 0
    assert out['Aeffgrat'] == 0
    assert len(out['R']) == 3
    assert np.isnan(out['R'][2])


def test_capture_res_aeff_nonfinite():
    '''Ensure that photons with nonfinite values will be ignored for resolving power
    but not for the effective area
    '''
    p = generate_test_photons(200)
    p['order'] = 0
    p['order'][50:] = 5
    p['xpos'] = np.nan

    resaeff = CaptureResAeff(A_geom=10., order_col='order',
                             orders=[0, 2, 5], dispersion_coord='xpos')
    out = resaeff(p)
    assert np.allclose(out['Aeff'], [2.5, 0., 7.5])
    assert out['Aeff0'] == 2.5
    assert out['Aeffgrat'] == 7.5
    assert len(out['R']) == 3
    assert np.isnan(out['R'][2])


def test_capture_res_aeff_CCDgap():
    '''Test the captures res/aeff class.

    Similar to the previous test, this is not a complete functional test,
    but it checks the interfaces.
    The actual function to calculate the effective area is tested elsewhere.
    '''
    p = generate_test_photons(200)
    p['order'] = 0
    p['order'][50:] = 5
    p['xpos'] = -100
    p['xpos'][50:] = np.random.normal(scale=1, size=150)
    p['CCD'] = 0
    # the last 100 photons don't hit a CCD and thus the Aeff is smaller
    p['CCD'][100:] = -1

    resaeff = CaptureResAeff_CCDgaps(A_geom=10., order_col='order',
                             orders=[0, 2, 5], dispersion_coord='xpos',
                             aeff_filter_col='CCD')
    out = resaeff(p)
    assert np.allclose(out['Aeff'], [2.5, 0., 2.5])
    assert np.isclose(out['Aeff0'], 2.5)
    assert np.isclose(out['Aeffgrat'], 2.5)
    assert len(out['R']) == 3
    assert np.isnan(out['R'][1])
    # Large rtol to reduce risk of random failures
    assert np.isclose(out['R'][2], 50., rtol=8)



# Licensed under GPL version 3 - see LICENSE.rst
from astropy.table import Table
import numpy as np
import pytest

from marxs.utils import DataFileFormatException, tablerows_to_2d

def test_tab_2d_OK():
    '''read a 3 * 2 table'''
    tab = Table(rows=[[1., 1., 0], [1., 2., 1],
                      [1.5, 1., 2], [1.5, 2., 3],
                      [1.8, 1., 4], [1.8, 2., 5]],
                names=['x', 'y','dat'])
    x, y, cols = tablerows_to_2d(tab)
    assert np.all(x == [1, 1.5, 1.8])
    assert np.all(y == [1, 2])
    assert np.all(cols[0] == np.arange(6).reshape((3, 2)))
    
    
def test_tab_2d_sizeerror():
    '''Like the first example, but the second row is missing.'''
    tab = Table(rows=[[1, 1, 1],
                      [1.5, 1., 3], [1.5, 2., 4.],
                      [1.8, 1., 5], [1.8, 2., 6.]],
                names=['x', 'y','dat'])
    with pytest.raises(DataFileFormatException, match='Data is not'):
        x, y, cols = tablerows_to_2d(tab)
            
    
def test_tab_2d_sorterror():
    '''Like the first example, but first and second entry are exchanged.'''
    tab = Table(rows=[[1, 2, 2],  [1, 1, 1],
                      [1.5, 1., 3], [1.5, 2., 4.],
                      [1.8, 1., 5], [1.8, 2., 6.]],
                names=['x', 'y','dat'])
    with pytest.raises(DataFileFormatException, match='not sorted'):
        x, y, cols = tablerows_to_2d(tab)


# Licensed under GPL version 3 - see LICENSE.rst

import numpy as np

class OrderColor():
    '''Consistent color scheme for plots that show multiple different orders

    Parameters
    ---------
    colormap : string
        Name of a matplotlib colormap
    max_order : int
        Maximum order that this color scheme will be used on (i.e. the end of
        the colorbar will be assigned ot this order)
    '''
    def __init__(self, colormap='nipy_spectral', max_order=15):
        import matplotlib.pyplot as plt
        cmap = plt.get_cmap(colormap)
        self.color = cmap(np.linspace(0, 1, abs(max_order + 1)))

    def __call__(self, order):
        kwargs = {}
        if order == 0:
            kwargs['color'] = 'k'
        else:
            kwargs['color'] = self.color[abs(order)]
        if order <= 0:
            kwargs['linestyle'] = '-'
        if order > 0:
            kwargs['linestyle'] = ':'
        if order == 0:
            kwargs['lw'] = 4
            kwargs['alpha'] = .7
        else:
            kwargs['lw'] = 2
        return kwargs



# Licensed under GPL version 3 - see LICENSE.rst
from copy import deepcopy
import numpy as np
import astropy.units as u
from astropy.table import QTable
from scipy.interpolate import interp1d
import transforms3d
import os

from astropy.table import Table
from scipy.interpolate import RectBivariateSpline

import marxs
from marxs.simulator import Sequence, KeepCol, ParallelCalculated
from marxs.optics import (GlobalEnergyFilter,
                          FlatDetector,
                          CircularDetector)
from marxs import optics
from marxs.design import rowland
import marxs.analysis
from marxs.math.utils import xyz2zxy
from marxs.utils import tablerows_to_2d
from marxs.design import tolerancing as tol
from marxs.missions.mitsnl.catgrating import catsupportbars, CATL1L2Stack, l1_order_selector
from marxs.missions.mitsnl.tolerances import wiggle_and_bar_tilt
from marxs.missions.athena.spo import SPOChannelMirror, ScatterPerChannel, spomounting
from marxs.missions.athena.spo import spogeom2pos4d
from marxs.missions.mitsnl.catgrating import InterpolateEfficiencyTable

from .ralfgrating import CATfromMechanical, CATWindow
from . import spo
from . import boom
from .load_csv import load_number, load_table
from .utils import tagversion, id_num_offset, config



__all__ = ['xyz2zxy',
           'Arcus', 'ArcusForPlot', 'ArcusForSIXTE']

defaultconf = rowland.double_rowland_from_channel_distance(750., 5915.51307, 12000. - 123.569239)
rowland.add_offset_double_rowland_channels(defaultconf)
for k, v in defaultconf.items():
    if isinstance(v, rowland.RowlandTorus):
        v.pos4d = xyz2zxy @ v.pos4d
for k, v in defaultconf['pos_opt_ax'].items():
    defaultconf['pos_opt_ax'][k] = xyz2zxy @ v

defaultconf['blazeang'] = 1.8
defaultconf['n_CCDs'] = 16
defaultconf['phi_det_start'] = 0.024  # was 0.037 at 600 mm channel spacing

# SPO scatter
# factor 2.3545 converts from FWHM to sigma
defaultconf['perpplanescatter'] = 1.5 / 2.3545 * u.arcsec
# 2 * 0.68 converts HPD to sigma
defaultconf['inplanescatter'] = 7. / (2 * 0.68) * u.arcsec

defaultconf['det_kwargs']= {
    'elem_class': FlatDetector,
    # orientation flips around CCDs so that det_x increases
    # with increasing x coordinate
    'elem_args': {'pixsize': 0.024, 'zoom': [1, 24.576, 12.288],
                  'orientation': np.array([[-1, 0, 0],
                                          [0, -1, 0],
                                          [0, 0, +1]])},
    'id_num_offset': 1,
}

# If we have the Arcus calibration database, load all the remaining
# configuration from there.
# If not (e.g. we are running in CI or readthedocs)
# do not load anything and use the defaults.
# Arcus specific tests will fail, but at least the model can be imported.
if 'data' in config:
    spogeom = load_table('spos', 'petallayout')
    spogeom['r_mid'] = (spogeom['outer_radius'] + spogeom['inner_radius']) / 2

    spo_pos4d = [np.dot(xyz2zxy, s) for s in spogeom2pos4d(spogeom)]

    defaultconf['spo_pos4d'] = spo_pos4d
    defaultconf['spo_geom'] = spogeom
    reflectivity = tablerows_to_2d(Table.read(os.path.join(config['data']['caldb_inputdata'],
                                                           'spos', 'coated_reflectivity.csv'),
                                              format='ascii.ecsv'))
    reflectivity_interpolator = RectBivariateSpline(reflectivity[0].to(u.keV),
                                                reflectivity[1].to(u.rad),
                                                reflectivity[2][0])

    defaultconf['reflectivity_interpolator'] = reflectivity_interpolator

    globalorderselector = InterpolateEfficiencyTable(
        Table.read(os.path.join(config['data']['caldb_inputdata'],
                 'gratings', 'efficiency.csv'), format='ascii.ecsv'))
    '''Global instance of an order selector to use in all CAT gratings.

    As long as the efficiency table is the same for all CAT gratings, it makes
    sense to define that globally. If every grating had its own independent
    order selector, we would have to read the selector file a few hundred times.
    '''


    defaultconf['gratinggrid'] = {'d_element': [32., 32.5],
                              'elem_class': CATL1L2Stack,
                              'elem_args': {'zoom': [1., 28./2, 28.5/2],
                                            'order_selector': globalorderselector,
                                            'l1_dims': {'bardepth': 0.0057 * u.mm,
                                                        'period': 0.005 * u.mm,
                                                        'barwidth': 0.0009 * u.mm},
                                            'l2_dims': {'bardepth': 0.5 * u.mm,
                                                        'period': 0.9622504 * u.mm,
                                                        'barwidth': 0.0962250 * u.mm},
                                            'qualityfactor': {'d': 200. * u.um,
                                                              'sigma': 1.75 * u.um},
                                            'l1_order_selector': l1_order_selector,
                              },
                              'parallel_spec': np.array([1., 0., 0., 0.]),
                   }


    ccdfwhm = QTable.read(os.path.join(config['data']['caldb_inputdata'],
                                       'detectors',
                                       'ccd_2021', 'arcus_ccd_rmf_20210211.txt'),
                         format='ascii.no_header', names=['energy', 'FWHM'])
    # Units currently not set in table
    ccdfwhm['energy'] *= u.keV
    ccdfwhm['FWHM'] *= u.eV
    defaultconf['ccdredist'] = ccdfwhm


# Other arguments for the detector - are they needed for something?
#defaultconf['detector']['d_element'] = [
#    defaultconf['det_kwargs']['elem_args']['zoom'][1] * 2 + 0.824 * 2 + 0.5,
#    defaultconf['det_kwargs']['elem_args']['zoom'][2] * 2 + 0.824 * 2 + 0.5,
#]


channels = ['1', '1m', '2', '2m']
# set of all channels. Used as default for a number of functions below.


def reformat_randall_errorbudget(budget, globalfac=0.8):
    '''Reformat the numbers from LSF-CAT-Alignment-v3.xls

    Randall gives 3 sigma errors, while I need 1 sigma a input here.
    Also, units need to be converted: mu -> mm, arcsec -> rad
    Last, global misalignment (that's not random) must be
    scaled in some way. Here, I use 0.8 sigma, which is the
    mean absolute deviation for a Gaussian.

    Parameters
    ----------
    budget : list
        See reference implementation for list format
    globalfac : ``None`` or float
        Factor to apply for global tolerances. A "global" tolerance is drawn
        only once per simulation. In contrast, for "individual" tolerances
        many draws are done and thus the resulting layout actually
        represents a distribution. For a "global" tolerance, the result hinges
        essentially on a single random draw. If this is set to ``None``,
        misalignments are drawn statistically. Instead, the tolerances can be
        scaled deterministically, e.g. by "0.8 sigma" (the mean absolute
        deviation for a Gaussian distribution).
    '''
    for row in budget:
        tol = np.array(row[2], dtype=float)
        tol[:3] = tol[:3] / 1000.  # mu to mm
        tol[3:] = np.deg2rad(tol[3:] / 3600.)  # arcsec to rad
        tol = tol / 3   # Randall gives 3 sigma values
        if row[1] == 'global':
            if globalfac is not None:
                tol *= globalfac
            else:
                tol *= np.random.randn(len(tol))

        row[3] = tol


class Aperture(optics.MultiAperture):
    def __init__(self, conf, channels=channels, **kwargs):
        # Set a little above entrance pos (the mirror) for display purposes.
        aper_z = 12200
        channels = deepcopy(channels)

        apers = []

        # Currently, channel 1 and 1m are so close, they need to
        # share an aperture
        #for pair in [('1', '1m', +1), ('2', '2m', -1)]:
        #    if (pair[0] in channels) and (pair[1] in channels):
        #        pos = [0, pair[2] * spo.rmid_spo(conf), aper_z]
        #        aperzoom = spo.zoom_to_cover_all_spos(conf, .8, 1.5)
        #        aperzoom[1] *=2
        #        rect = optics.RectangleAperture(position=pos,
        #                                        zoom=aperzoom,
        #                                        orientation=xyz2zxy[:3, :3])
        #        rect.display['outer_factor'] = 2.
        #        apers.append(rect)
        #        channels.remove(pair[0])
        #        channels.remove(pair[1])

        # Now, add individual apertures if they are not double
        # Thus, needs to be geometrically bigger for off-axis sources.
        aperzoom = spo.zoom_to_cover_all_spos(conf, 1.1, 0.8)
        for chan in channels:
            pos = conf['pos_opt_ax'][chan][:3].copy()
            pos[2] += aper_z
            if '1' in chan:
                pos[1] += spo.rmid_spo(conf)
            elif '2' in chan:
                pos[1] -= spo.rmid_spo(conf)
            else:
                raise ValueError('No rules for channel {}'.format(chan))
            pos[2] = aper_z

            rect = optics.RectangleAperture(position=pos,
                                            zoom=aperzoom,
                                            orientation=xyz2zxy[:3, :3])
            rect.display['outer_factor'] = 1.3
            apers.append(rect)

        super(Aperture, self).__init__(elements=apers, **kwargs)


class SimpleSPOs(Sequence):

    def __init__(self, conf, channels=channels,
                 **kwargs):
        rot180 = transforms3d.euler.euler2mat(np.pi, 0, 0, 'szyx')
        # Make lens a little larger than aperture, otherwise an non on-axis ray
        # (from pointing jitter or an off-axis source) might miss the mirror.
        mirror = []
        for chan in channels:
            entrancepos = conf['pos_opt_ax'][chan][:3].copy()
            entrancepos[2] += 12000
            rot = np.eye(3) if '1' in chan else rot180
            mirror.append(SPOChannelMirror(position=entrancepos,
                                           orientation=rot,
                                           reflectivity_interpolator=conf['reflectivity_interpolator'],
                                           spo_pos4d=conf['spo_pos4d'],
                                           spo_geom=conf['spo_geom'],
                                           id_num_offset=id_num_offset[chan]))
            mirror.append(ScatterPerChannel(position=entrancepos,
                                            min_id=id_num_offset[chan],
                                            max_id=id_num_offset[chan] + 1000,
                                            inplanescatter=conf['inplanescatter'],
                                            perpplanescatter=conf['perpplanescatter'],
                                            orientation=xyz2zxy[:3, :3]))
        mirror.append(spomounting)
        self.geometricopening = load_number('spos', 'geometricthroughput', 'transmission') * \
            load_number('spos', 'porespecifications', 'transmission')
        mirror.append(GlobalEnergyFilter(filterfunc=lambda e: self.geometricopening,
                                         name='SPOgeometricthrougput'))

        tab = load_table('spos', 'lossterm')
        en = tab['energy'].to(u.keV, equivalencies=u.spectral())
        mirror.append(GlobalEnergyFilter(filterfunc=interp1d(en, tab[tab.colnames[1]]),
                                         name='Loss_from_nonideal_spos'))

        super(SimpleSPOs, self).__init__(elements=mirror, **kwargs)


class CATGratings(Sequence):
    def __init__(self, conf, channels=channels, **kwargs):

        elements = []

        for chan in channels:
            shift_matrix = np.eye(4)
            shift_matrix[:, 3] = conf['pos_opt_ax'][chan][:]

            elements.append(CATfromMechanical(pos4d=shift_matrix,
                                              channel=chan, conf=conf,
                                              id_num_offset=id_num_offset[chan],
                                              ))
        elements.append(catsupportbars)
        super(CATGratings, self).__init__(elements=elements, **kwargs)


class FiltersAndQE(Sequence):
    '''
    Parameters
    ----------
    conf, channels : any
        Parameters are accepted for consistency with other elements in
        Arcus, but currently, they are not needed here because all relevant
        settings are hardcoded.
    kwargs_interp1d : dict
        keywords for the filter function, for example to allow values outside
        the range available in the configuration files:
        `{'bounds_error'=False}`.
    '''

    filterlist = [('filters', 'sifilter'),
                  ('filters', 'opticalblocking'),
                  ('filters', 'uvblocking'),
                  ('detectors', 'contam'),
                  ('detectors', 'qe')]

    def get_filter(self, directory, name):
        tab = load_table(directory, name)
        en = tab['energy'].to(u.keV, equivalencies=u.spectral())
        return GlobalEnergyFilter(filterfunc=interp1d(en, tab[tab.colnames[1]],
                                                      **self.kwargs_interp1d),
                                  name=name)

    def __init__(self, conf=None, channels=None, kwargs_interp1d={}, **kwargs):
        self.kwargs_interp1d = kwargs_interp1d
        elems = [self.get_filter(*n) for n in self.filterlist]
        super(FiltersAndQE, self).__init__(elements=elems, **kwargs)


class DetMany(rowland.RectangularGrid):

    y_range=[-400, 400]

    def __init__(self, conf, **kwargs):
        super(DetMany, self).__init__(rowland=conf['rowland_detector'],
                                      elem_class=self.elem_class,
                                      elem_args=self.elem_args,
                                      d_element=self.d_element,
                                      y_range=self.y_range,
                                      # convention is to start counting at 1
                                      id_num_offset=1,
                                      id_col='CCD',
                                      guess_distance=-25.)


class Det16(DetMany):
    '''Place only hand-selected 16 CCDs'''
    def __init__(self, conf, theta=[3.1255, 3.1853, 3.2416, 3.301], **kwargs):
        self.theta = theta
        super(Det16, self).__init__(conf, **kwargs)
        assert len(self.elements) == 16
        # but make real detectors orange
        disp = deepcopy(self.elements[0].display)
        disp['color'] = 'orange'
        for e in self.elements:
            e.display = disp


class DetCamera(ParallelCalculated):
    '''CCD detectors in camera layout

    8 CCDs in each strip, with gaps between them in 3-2-3 groups
    to match the shadows from the filter support.
    One strip is offset by a few mm so that the chip gaps fall
    into slightly different positions.
    '''
    d_fsupport = 4.
    '''Distance between CCDs under a filter support bar in mm'''

    d_ccd = 2.15
    '''Minimal distance between CCDs in mm'''

    offset = [-5., 5.]
    '''offset of one strip vs the other to avoid matching chip gaps in mm'''

    id_col = 'CCD'

    def __init__(self, conf):
        r = conf['rowland_detector'].r
        phi_m = np.arcsin(conf['d'] / r) + np.pi
        ccd = conf['det_kwargs']['elem_args']['zoom'][1]
        p0 = conf['phi_det_start']
        gaps = np.array([0, self.d_ccd, self.d_fsupport, self.d_ccd,
                         self.d_ccd, self.d_fsupport,
                         self.d_ccd, self.d_ccd])
        theta1 = (p0 + ccd / r) + np.arange(8) * 2 * ccd / r + gaps.cumsum() / r
        theta = np.hstack([phi_m - theta1 - self.offset[0] / r,
                           phi_m + self.offset[1] / r + theta1])
        # Sort so that CCD number increases from -x to +x
        theta.sort()
        theta = theta[::-1]
        pos = conf['rowland_detector'].parametric(theta, 0)

        super().__init__(pos_spec=pos,
                         normal_spec=conf['rowland_detector'].normal,
                         parallel_spec=conf['rowland_detector'].pos4d @ np.array([0, 1, 0, 0]),
                         **conf['det_kwargs'],
                         )
        assert len(self.elements) == conf['n_CCDs']
        # but make real detectors orange
        disp = deepcopy(self.elements[0].display)
        disp['color'] = 'orange'
        for e in self.elements:
            e.display = disp


class FocalPlaneDet(marxs.optics.FlatDetector):
    loc_coos_name = ['detfp_x', 'detfp_y']
    detpix_name = ['detfppix_x', 'detfppix_y']

    def __init__(self, **kwargs):
        if ('zoom' not in kwargs) and ('pos4d' not in kwargs):
            kwargs['zoom'] = [.2, 10000, 10000]
        if ('orientation' not in kwargs) and ('pos4d' not in kwargs):
            kwargs['orientation'] = xyz2zxy[:3, :3]
        super(FocalPlaneDet, self).__init__(**kwargs)


class PerfectArcus(Sequence):
    '''Default Definition of Arcus without any misalignments'''
    aper_class = Aperture
    spo_class = SimpleSPOs
    #gratings_class = RegularGrid
    gratings_class = CATGratings
    filter_and_qe_class = FiltersAndQE
    '''Set any of these classes to None to not have them included.
    (e.g. SIXTE does filters and QE itself).
    '''

    list_of_classes = ['aper_class', 'spo_class', 'gratings_class',
                       'filter_and_qe_class']

    def add_boom(self, conf):
        '''Add four sided boom. Only the top two bays contribute any
        absorption, so we can save time by not modelling the remaining bays.'''

        return []
        #return [boom.FourSidedBoom(orientation=xyz2zxy[:3, :3],
        #                           position=[0, 0, 596.],
        #                           boom_dimensions={'start_bay': 6})]

    def add_detectors(self, conf):
        '''Add detectors to the element list

        This is a separate function that is called from __init__
        because all detectors need different parameters. Placing this
        specific code in it's own function makes it easy to override
        for derived classes.

        '''
        # rotate such that phi=0 is at the bottom
        rot = transforms3d.axangles.axangle2mat(np.array([0, 1, 0]), np.pi)
        circdet = CircularDetector(orientation=xyz2zxy[:3, :3] @ rot,
                                   zoom=[defaultconf['rowland_detector'].r,
                                         defaultconf['rowland_detector'].r,
                                         20],
                                   position=[0, 0, np.sqrt(defaultconf['rowland_detector'].r**2 - conf['d']**2)],
                                   )
        circdet.display['opacity'] = 0.1
        circdet.detpix_name = ['circpix_x', 'circpix_y']
        circdet.loc_coos_name = ['circ_phi', 'circ_y']
        circproj = marxs.analysis.ProjectOntoPlane(orientation=xyz2zxy[:3, :3])
        circproj.loc_coos_name = ['projcirc_x', 'projcirc_y']
        reset = marxs.simulator.simulator.Propagator(distance=-100.)
        twostrips = DetCamera(conf)
        proj = marxs.analysis.ProjectOntoPlane(orientation=xyz2zxy[:3, :3])
        detfp = FocalPlaneDet()
        return [circdet, circproj, reset, twostrips, proj, detfp]

    def post_process(self):
        self.KeepPos = KeepCol('pos')
        return [self.KeepPos]

    def __init__(self, conf=defaultconf, channels=channels,
                 **kwargs):
        elem = []
        for c in self.list_of_classes:
            cl = getattr(self, c)
            if cl is not None:
                elem.append(cl(conf, channels))
        elem.extend(self.add_boom(conf))
        elem.extend(self.add_detectors(conf))
        elem.append(tagversion)
        super(PerfectArcus, self).__init__(elements=elem,
                                           postprocess_steps=self.post_process(),
                                           **kwargs)


class Arcus(PerfectArcus):
    def __init__(self, conf=defaultconf, channels=channels,
                 **kwargs):
        super(Arcus, self).__init__(conf=conf, channels=channels, **kwargs)
        for row in conf['alignmentbudget']:
            elem = self.elements_of_class(row[0])
            if row[1] == 'global':
                tol.moveglobal(elem, *row[3])
            elif row[1] == 'individual':
                tol.wiggle(elem, *row[3])
            elif row[1] == 'individual_with_tilt':
                wiggle_and_bar_tilt(elem, *row[3])
            else:
                raise NotImplementedError('Alignment error {} not implemented'.format(row[1]))


class ArcusForPlot(PerfectArcus):
    '''Arcus with setting that are good for 3D plots

    In particular, there is a full boom and no large catch-all focal plane.
    '''
    def add_boom(self, conf):
        return [boom.FourSidedBoom(orientation=xyz2zxy[:3, :3],
                                   position=[0, 0, 546.])]

    def add_detectors(self, conf):
        '''Add detectors to the element list

        This is a separate function that is called from __init__
        because all detectors need different parameters. Placing this
        specific code in it's own function makes it easy to override
        for derived classes.

        '''
        return [DetCamera(conf)]


class ArcusForSIXTE(Arcus):
    filter_and_qe_class = None

    def add_detectors(self, conf):
        return [FocalPlaneDet()]

# Note: We are inventing a new format here.
# - individual -> wiggle
# - global -> moveglobal
# When revising, might be better to put functions directly in there to
# avoid confusion.

align_requirement_smith = [
    [SPOChannelMirror, 'individual', [20, 100, 500, 300, 300, 10],
     None, 'individual SPO in petal'],
    # The following term is for the alignment of each channel
    # (SPO + CAT petal) relative to the front assembly, i.e. SPO and
    # CAT petal are moved together.
    # To make this work, MARXS needs relative position uncertainties
    # which, at present, are not implemented.
    # However, these numbers are smaller than "Camera to front assembly"
    # and for simulations of just one channel would have the same effect.
    # Thus, neglecting them for now is a only a small contribution.
    #[spo.SPOChannelMirror, 'global', [2000., 800, 800, 180, 180, 180],
    # None, 'SPO petal to front assembly'],
    [CATfromMechanical, 'global', [1000, 1000, 1000, 300, 300, 600],
     None, 'CAT petal to SPO petal'],
    [CATfromMechanical, 'individual', [1000, 1000, 200, 300., 180, 300],
     None, 'CAT windows to CAT petal'],
    [CATWindow, 'individual', [1000, 1000, 200, 300, 180, 300],
     None, 'individual CAT to window'],
    [DetCamera, 'global', [5000, 2000, 1000, 180, 180, 180],
     None, 'Camera to front assembly']]
'''This is taken from LSF-CAT-Alignment-v17.xls from R. Smith'''

align_requirement = deepcopy(align_requirement_smith)
reformat_randall_errorbudget(align_requirement)

defaultconf['alignmentbudget'] = align_requirement



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from astropy.table import Table

from ..optics import FlatOpticalElement
from ..math.utils import h2e


__all__ = ['ProjectOntoPlane', 'facet_table']


class ProjectOntoPlane(FlatOpticalElement):
    '''Project photon positions onto a plane.

    *Projection* as used in this class differs from propagating the photons
    until they intersect with a plane. Instead, the current photon positions
    are projected parallel to the normal of the plane.

    Note that the default output columns are called proj_x and _y independent
    of the actual orientation of the plane.

    Examples
    --------
    Project some points onto a plane. If no orientation is given to the
    ``ProjectionOntoPlane`` object, the default is to place it in the yz-plane.

    >>> from astropy.table import Table
    >>> import numpy as np
    >>> from marxs.analysis.coords import ProjectOntoPlane
    >>> photons = Table()
    >>> photons['pos'] = np.array([[3, 3, 3, 1], [12, -1, 0, 1]])
    >>> yzplane = ProjectOntoPlane()
    >>> photons = yzplane(photons)
    >>> photons['proj_x'].data
    array([ 3., -1.])
    >>> photons['proj_y'].data
    array([3., 0.])
    '''

    loc_coos_name = ['proj_x', 'proj_y']
    '''name for output columns of the projected position in plane coordinates.'''

    display = {'shape': 'None'}

    def __call__(self, photons):
        vec_center_inter = - h2e(self.geometry['center']) + h2e(photons['pos'])
        photons[self.loc_coos_name[0]] = np.dot(vec_center_inter, h2e(self.geometry['e_y']))
        photons[self.loc_coos_name[1]] = np.dot(vec_center_inter, h2e(self.geometry['e_z']))
        return photons


def facet_table(container, project_plane=ProjectOntoPlane()):
    '''Get table of facet properties

    Parameters
    ----------
    container : `marxs.simulator.Parallel` instance
        This can be some kind of container, typically a grating array
        on the Rowland torus.
    project_plane : `marxs.analysis.ProjectOntoPlane`
        radius and angle are calculated as projected onto this plane

    Returns
    -------
    facettab : `astropy.table.Table`
        Table with facet properties.
        All output columns are prevaced with "facet_" to avoid a clash with columns likely
        to be found in photon tables. That enables easy merging of facet tables with photon tables
        to add the properties of a facet that the photon passed through into the photon table
        like so:

            >>> from astropy import table
            >>> table.join(photons, facettab, join_type='left')  # doctest: +SKIP
    '''
    project_plane.loc_coos_name = ['facet_projx', 'facet_projy']

    facetpos = h2e(np.stack(container.elem_pos)[:, :, 3])
    tab = Table({'pos': np.stack(container.elem_pos)[:, :, 3]})
    tab = project_plane(tab)

    tab['facet_ang'] = np.arctan2(tab['facet_projy'], tab['facet_projx'])
    # projx and proy are measured from center of plane, not from origin of cooridnate system!
    tab['facet_rad'] = np.sqrt(tab['facet_projy']**2 + tab['facet_projx']**2)
    tab['facet'] = [e.id_num for e in container.elements]
    for i, s in enumerate('xyz'):
        tab[f'facet_{s}'] = facetpos[:, i]
    tab.remove_column('pos')

    return tab



# Licensed under GPL version 3 - see LICENSE.rst
import re
import datetime

import numpy as np
from scipy import interpolate
from astropy.table import Column, Table
import astropy.units as u

# We want Sherpa to be an optional dependency, so the imports happen in
# individual functions, not here.
# from sherpa.models import NormGauss1D, Scale1D
# from sherpa.astro.models import Lorentz1D
# from sherpa import stats, optmethods
# from sherpa.fit import Fit
# from sherpa.data import Data1DInt


from .ogip import RMF

COPY_LSFPARM_KEYWORDS = [
    # OGIP mantatory
    'TELESCOP', 'INSTRUME', 'FILTER',
    # Chandra
    'DETNAM', 'RAND_TG', 'GRATING', 'GRATTYPE', 'TG_M', 'ORDER', 'SHELL',
]

modelnames = re.compile('(GAUSS|LORENTZ)[0-9]+_PARMS')
gaussn = re.compile('GAUSS(?P<n>[0-9]+)_PARMS')
lorentzn = re.compile('LORENTZ(?P<n>[0-9]+)_PARMS')

# convert parameters from CALDB convention to Sherpa
# For NormGauss1d Sherpa uses FWHM, while CALDB uses sigma
gconv = np.array([1, 2 * np.sqrt(2 * np.log(2)), 1])
lconv = np.array([1, 1, 1])


# Is this used anywhere? Usually, do it from spline.
# Can always convert CALDB to spline and then use that.
# So I think this is just a second implementation of the same thing
# which should be be here because now I have to debug etc. twice.
def caldb2sherpa(caldb, row, iwidth, iwave):
    '''Convert CALDB entries to Sherpa model

    Convert entries in a CALDB lsfparm files to a Sherpa model.  The
    lsfparm files can contain several rows, for different off-axis
    angles and radii and in each row there will be entries for a
    number of wavelength points and extraction width.

    This function expects as input the index numbers for row,
    wavelength as index for the wavelength array etc. In practical
    applications, the CALDB file will be queried for a specific
    position, wavelength etc., but for development purposes it is
    useful to go into the raw array, e.g. to read some unrelated CALDB
    file (say for a different detector) to use as a starting point to
    fit the lsfparm parameters or to plot different settings for
    comparison.

    caldb : `astropy.table.Table`
        CALDB lsfparm table

    '''
    from sherpa.astro.models import Lorentz1D
    from sherpa.models import NormGauss1D, Scale1D

    model = []
    for col in caldb.colnames:
        if col == 'EE_FRACS':
            eef = Scale1D(name='EE_FRACS')
            eef.c0 = caldb['EE_FRACS'][row][iwidth, iwave]
            # model is underdetermined if this the ampl of all functions
            # is left free
            eef.c0.frozen = True
        elif gaussn.match(col):
            newg = NormGauss1D(name=col)
            newg.ampl, newg.fwhm, newg.pos = caldb[col][row][iwidth, iwave, :] * gconv
            model.append(newg)
        elif lorentzn.match(col):
            newg = Lorentz1D(name=col)
            newg.ampl, newg.fwhm, newg.pos = caldb[col][row][iwidth, iwave, :] * lconv
            model.append(newg)

    sumampl = np.sum([m.ampl.val for m in model if isinstance(m, NormGauss1D) or isinstance(m, Lorentz1D)])
    for m in model:
        if isinstance(m, NormGauss1D) or isinstance(m, Lorentz1D):
            m.ampl.val = m.ampl.val / sumampl
    # Start value is 0, unless we explicitly set the start value. So, split models and pass [0]
    # as start value to avoid a numerical 0.0 in the model expression.
    return eef * sum(model[1:], model[0])


def flatten_sherpa_model(model):
    if hasattr(model, 'parts'):
        modellist = []
        for p in model.parts:
            modellist.extend(flatten_sherpa_model(p))
        return modellist
    else:
        return [model]


def sherpa2caldb(shmodel):
    '''
    shmodel : Sherpa model instance
    '''
    from sherpa.astro.models import Lorentz1D
    from sherpa.models import NormGauss1D

    d = {}
    for comp in flatten_sherpa_model(shmodel):
        if isinstance(comp, NormGauss1D):
            modelpars = np.array([comp.ampl.val, comp.fwhm.val,
                                  comp.pos.val]) / gconv
        elif isinstance(comp, Lorentz1D):
            modelpars = np.array([comp.ampl.val, comp.fwhm.val,
                                  comp.pos.val]) / lconv
        else:
            raise Exception(f'Component {comp} not valid for LSFPARM files')
        d[comp.name] = modelpars
    return d


def empty_lsfparmtable(widths, waves, model, extname, order):
    '''
    Parameters
    ----------
    width : np.array
        extraction width
    waves
    model : Sherpa model instance
        This instance is inspected for the name of the model components to
        set up the correct columns
    '''
    from sherpa.astro.models import Lorentz1D
    from sherpa.models import NormGauss1D

    tab = Table()
    tab.add_column(Column(data=[len(widths)], name='NUM_WIDTHS',
                          dtype=np.int32))
    tab.add_column(Column(data=[widths], name='WIDTHS', dtype=np.float32,
                          shape=(len(widths)), unit=u.degree))
    tab.add_column(Column(data=[len(waves)], name='NUM_LAMBDAS',
                          dtype=np.int32))
    tab.add_column(Column(data=[waves], name='LAMBDAS',
                          dtype=np.float32, shape=(len(waves)),
                          unit=u.Angstrom))
    # From the ICD: The columns TG LAM LO and TG LAM HI give that lower and
    # upper wavelength of
    # the box that was used to extract the LSF data.
    # I think these columns are informational only, but I need to check that.
    dwave = np.max(np.diff(waves)) + 1.2
    tab.add_column(Column(data=[waves - dwave], name='TG_LAM_LO',
                          dtype=np.float32, shape=(len(waves)),
                          unit=u.Angstrom))
    tab.add_column(Column(data=[waves + dwave], name='TG_LAM_HI',
                          dtype=np.float32, shape=(len(waves)),
                          unit=u.Angstrom))
    tab.add_column(Column(name='EE_FRACS', length=1, dtype=np.float32,
                          shape=(len(widths), len(waves))))
    # See which columns are needed to describe the models
    modelcomps = flatten_sherpa_model(model)
    compnames = [m.name for m in modelcomps]
    # Check names are unique and follow DPID naming scheme
    if not len(set(compnames)) == len(compnames):
        raise ValueError("Names of model components are not unique (using default names?)")
    # Check models are right type and follow DPID naming rules:
    for m in modelcomps:
        if not isinstance(m, (NormGauss1D, Lorentz1D)):
            raise ValueError('Model can only contain Sherpa NormGauss1D and Lorentz1D components.')
    # Check model and name match
    # Check count starts at 1 (not 0), consequtive numbers
    for m in compnames:
        tab.add_column(Column(name=m + '_PARMS', length=1, dtype=np.float32,
                              shape=(len(widths), len(waves), 3)))

    # Add header keywords.
    # TODO: Not done yet, just adding what I can to make mkrmf read it
    # Not sure for all keywords what they mean or should be
    tab.meta['EXTNAME'] = extname
    tab.meta['CCLS0001'] = 'CPF'
    tab.meta['CDTP0001'] = 'DATA'
    tab.meta['CCNM0001'] = 'LSFPARM'
    tab.meta['RAND_TG'] = '0.0'   # make this a parameter
    tab.meta['LONGSTRN'] = 'OGIP 1.0'
    tab.meta['CREATOR'] = 'lsfparm.py'  # update
    tab.meta['DATE'] = str(datetime.datetime.now()).replace(' ', 'T')[:19],
    tab.meta['CONTENT'] = 'CDB_' + tab.meta['EXTNAME'] + '_LSFPARM'
    tab.meta['HDUCLASS'] = 'ASC',
    tab.meta['HDUCLAS1'] = 'PARAMETERS'
    tab.meta['HDUCLAS2'] = 'PSF'
    tab.meta['HDUCLAS3'] = 'LSF'
    tab.meta['TG_M'] = order  # not sure what goes here. Check existing files!
    tab.meta['ORDER'] = order
    tab.meta['DETNAM'] = 'UNKNOWN'
    tab.meta['INSTRUME'] = 'UNKNOWN'
    tab.meta['CVSD0001'] = '1999-07-22T00:00:00'  # Chandra as an example
    tab.meta['CVST0001'] = '00:00:00'
    tab.meta['TELESCOP'] = 'UNKOWN'
    tab.meta['FILTER'] = 'UNKNOWN'
    return tab


def lsfparmtable_add_Chandra_header(tab, evt):
    tab.meta['CDES0001'] = tab.meta['EXTNAME'] + \
        ' line spread function, input for mkgrmf'
    tab.meta['DETNAM'] = evt.meta['DetectorType']
    tab.meta['INSTRUME'] = evt.meta['DetectorType'].split('-')[0]
    tab.meta['CVSD0001'] = '1999-07-22T00:00:00'
    tab.meta['CVST0001'] = '00:00:00'
    tab.meta['TELESCOP'] = 'CHANDRA'
    tab.meta['FILTER'] = 'NONE'
    tab.meta['GRATING'] = evt.meta['GratingType']
    tab.meta['GRATTYPE'] = tab.meta['EXTNAME'] if evt.meta['GratingType']=='HETG' else ''  # Check value for LETG

    tab.meta['SHELL'] = {'MEG': '1100', 'HEG': '0011', 'LEG': '1111'}[tab.meta['EXTNAME']]  # put right name here for LEG
    tab.meta['CBD10001'] = 'ORDER({})'.format(tab.meta['ORDER'])
    tab.meta['CBD20001'] = 'RAND_TG({})'.format(tab.meta['RAND_TG'])
    tab.meta['CBD30001'] = 'SHELL({})'.format(tab.meta['SHELL'])
    tab.meta['CBO10001'] = 'GRATING({})'.format(tab.meta['GRATING'])
    tab.meta['CBO20001'] = 'GRATTYPE({})'.format(tab.meta['GRATTYPE'])  # check LEG if present at all
    tab.meta['CBO30001'] = 'TG_M({})'.format(tab.meta['TG_M'])
    # No idea what those mean...
    #('FDLT0001', 2.0),
    #('CAL_QUAL', 0),

    return tab


class interp1d_2dsignature(interpolate.interp1d):
    '''1D interp with the same signature as `scipy.interpolate.interp2d`

    For use with LSFPARM files that have only one WIDTH.

    '''
    def __init__(self, x, y, z, **kwargs):
        super().__init__(y, z, **kwargs)

    def __call__(self, x, y):
        return super().__call__(y)


def CALDB_interp(row, kind='cubic', **kwargs):
    '''Get interpolating functions from values in LSFPARM

    Parameters
    ----------
    row : `astropy.table.Row`
        Table row from CALDB LSFPARM file
    **kwargs
        All other keywords arguments are passed to
        `scipy.interpolate.interp1d` or `scipy.interpolate.interp2d`.
    '''
    interpolators = {}

    if row['NUM_WIDTHS'] > 1:
        interp = interpolate.interp2d
    else:
        interp = interp1d_2dsignature

    for col in row.colnames:
        if col == 'EE_FRACS':
            interpolators[col] = interp(row['WIDTHS'],
                                        row['LAMBDAS'], row[col],
                                        kind=kind,
                                        **kwargs)
        elif modelnames.match(col):
            interpolators[col] = [interp(row['WIDTHS'],
                                         row['LAMBDAS'],
                                         row[col][:, :, i].squeeze(),
                                         kind=kind,
                                         **kwargs)
                                  for i in [0, 1, 2]]
    return interpolators


def sherpa_from_spline(splines, width, wave):
    '''
    To-Do: Convert to correct unit before calling .value
    To-Do: np.abs is really just there if an interpolation
      goes below 0. Setting to 0 or linearly interpolating
      would be better, but I simply want a fast solution right now.
    '''
    from sherpa.models import NormGauss1D, Scale1D
    from sherpa.astro.models import Lorentz1D

    model = []
    for col in splines:
        if col == 'EE_FRACS':
            eef = Scale1D(name='EE_FRACS')
            eef.c0 = splines[col](width.value, wave.value)
            # model is underdetermined if this the ampl of all functions
            # is left free
            eef.c0.frozen = True
        elif gaussn.match(col) or lorentzn.match(col):
            if gaussn.match(col):
                new = NormGauss1D(name=col)
                conv = gconv
            else:
                new = Lorentz1D(name=col)
                conv = lconv
            new.ampl, new.fwhm, new.pos = \
                np.abs(np.stack([splines[col][i](width.value, wave.value)
                          for i in [0, 1, 2]]).flatten() * conv)
            model.append(new)

    sumampl = np.sum([m.ampl.val for m in model
                      if isinstance(m, NormGauss1D) or
                      isinstance(m, Lorentz1D)])
    for m in model:
        if isinstance(m, NormGauss1D) or isinstance(m, Lorentz1D):
            m.ampl.val = m.ampl.val / sumampl
    # Start value is 0, unless we explicitly set the start value.
    # So, split models and pass [0]
    # as start value to avoid a numerical 0.0 in the model expression.
    return eef * sum(model[1:], model[0])


def make_rmf(lsfparmrow, wave_edges, width, threshold=1e-6, kw_interp={}):
    '''
    makes PHA channels (because that's what the lsfparm files are made for)
    I made a choice here: writing with TLMIN=0, which is valid OGIP
    '''
    ebounds = RMF.ebounds_from_edges(wave_edges)
    matrix = Table(names=['ENERG_LO', 'ENERG_HI', 'N_GRP',
                          'F_CHAN', 'N_CHAN', 'MATRIX'],
                   data=[np.zeros(len(ebounds), dtype=d) for d in [
                       np.float32, np.float32, np.int16,
                       object, object, object]],
                   units=[u.keV, u.keV, None, None, None, None])

    matrix['ENERG_LO'] = ebounds['E_MIN']
    matrix['ENERG_HI'] = ebounds['E_MAX']
    matrix.meta['DETCHANS'] = len(wave_edges) - 1
    matrix.meta['LO_THRES'] = threshold
    matrix.meta['TLMIN{}'.format(matrix.colnames.index('F_CHAN') + 1)] = 0
    ebounds.meta['TLMIN{}'.format(ebounds.colnames.index('CHANNEL') + 1)] = 0
    for tab in [ebounds, matrix]:
        for k in COPY_LSFPARM_KEYWORDS:
            if k in lsfparmrow.meta:
                tab.meta[k] = lsfparmrow.meta[k]

    ang_lo = matrix['ENERG_HI'].to(u.Angstrom,
                                   equivalencies=u.spectral())
    ang_hi = matrix['ENERG_LO'].to(u.Angstrom,
                                   equivalencies=u.spectral())
    midwave = 0.5 * (ang_lo + ang_hi)

    splines = CALDB_interp(lsfparmrow, **kw_interp)
    for r, wav in enumerate(midwave):
        func = sherpa_from_spline(splines, width, wav)
        # We want sherpa in increasing wavelength, so reverse order
        fullmatrix = func(ang_lo[::-1], ang_hi[::-1])
        # but then we want the RMF in increasing energy, so we reverse again
        out = RMF.arr_to_rmf_matrix_row(fullmatrix[::-1], 0,
                                        threshold=threshold)
        for i, col in enumerate(['N_GRP', 'F_CHAN', 'N_CHAN', 'MATRIX']):
            matrix[col][r] = out[i]
    return RMF(matrix, ebounds, CHANTYPE='PHA', HDUCLAS3='REDIST')


def fit_LSF(evt, model, wavebin=0.001, d_wave=0.03, colname='tg_mlam',
                stat=None, method=None):
    '''Fit an LSF model to a simulated set of photons

    Photons are binned into a histogram and the model is fit to that.
    Since the histogram is constructed using the "probability" column
    in the event list, if will contain non-integer data, which does not
    follow a Poisson distribution.
    Unfortunately, the statistical error in each bin is not well defined,
    but some number is needed for the fit. So, the square root of the
    number of photons in each bin (with a floor of 1) is used.
    This is not ideal, but it gives satisfactory results for typical cases.

    evt : `astropy.table.Table`
        Table with photons. Must have columns 'probability' and 'colname'.
        This table must already be filtered to contain only photons
        that are part of the LSF (e.g. filter on order and wavelength)
    model : `sherpa.models.model.Model`
        LSF model to be fit.
    wavebin : float
        Wavelength bin size for histogram
    d_wave : float
        Half-width of wavelength range to fit
    colname : str
        Name of column in evt that contains the wavelength. The default
        name is taken from the Chandra terminaology.
    stat : `sherpa.stats.Stat`
        If None, use `sherpa.stats.LeastSq`
    method : `sherpa.optmethods.OptMethod`
        If None, use `sherpa.optmethods.LevMar`

    Returns
    -------
    sdata : `sherpa.data.Data1D`
        Data object with the histogram
    res : `sherpa.fit.FitResults`
        Results of the fit
    '''
    from sherpa.fit import Fit
    from sherpa.data import Data1DInt
    from sherpa import stats, optmethods

    if stat is None:
        stat = stats.LeastSq()
    if method is None:
        method = optmethods.LevMar()

    hist, edges = np.histogram(evt[colname].value, weights=evt['probability'],
                               bins=evt[colname].value.mean() + np.arange(- d_wave, d_wave, wavebin))
    # Get error with 0 in the uncertainties
    # So, just setting to 1 for now. Should fix properly
    # e.g. reduce binsize to prevent that or use other statistic and not use that
    sdata = Data1DInt('counts_histogram', edges[:-1], edges[1:], hist,
                      staterror=np.clip(np.sqrt(hist), 1, np.inf))
    modfit = Fit(sdata, model, stat=stat, method=method)
    res = modfit.fit()
    # Check if fit was successful
    return sdata, res


def plot_LSFfit(sdata, model, axes):
    '''Plot an LSF fit for visual inspection

    Parameters
    ----------
    sdata : `sherpa.data.Data1D`
        Data object with the histogram
    model : `sherpa.models.model.Model`
        LSF model from the fit to be plotted
    axes : list of `matplotlib.axes.Axes`
        The first axis is used linar plot, highligting the fit to the
        line center and the second axis for a log plot, highlighting the
        wings.
    '''
    for ax in axes:
        ax.plot(sdata.x, sdata.y, 'k', label='Data')
        ax.plot(sdata.x, model(sdata.xlo, sdata.xhi), linewidth=2, label='model')
        for m in flatten_sherpa_model(model):
            ax.plot(sdata.x, m(sdata.xlo, sdata.xhi), linewidth=2, label=m.name)
        ax.set_xlabel('wavelength [Ang]')
    wave = np.mean(sdata.x)  # Fitting regions are centered on wavelength
    d_wave = (np.max(sdata.x) - np.min(sdata.x)) / 2
    axes[1].set_yscale('log')
    axes[1].set_ylim([1, np.max(sdata.y)])



# Licensed under GPL version 3 - see LICENSE.rst
from astropy.table import Table
from astropy.utils.data import get_pkg_data_filename

from marxs.missions.mitsnl.catgrating import InterpolateEfficiencyTable as IET

# CAT gratings tabulated data
order_selector_Si = IET(Table.read(
    get_pkg_data_filename('data/Si_efficiency_5_7.dat',
                          package='marxs.missions.mitsnl'),
                          format='ascii.ecsv'))
order_selector_Si.coating = 'None'
order_selector_Pt = IET(Table.read(
    get_pkg_data_filename('data/SiPt_efficiency_5_7.dat',
                          package='marxs.missions.mitsnl'),
                          format='ascii.ecsv'))
order_selector_Pt.coating = 'Pt'



# Licensed under GPL version 3 - see LICENSE.rst
import numpy as np
from transforms3d.axangles import axangle2mat
from transforms3d.affines import compose

from marxs.design.rowland import RectangularGrid
from marxs.simulator import Parallel, Sequence
from marxs.missions.mitsnl.catgrating import (CATL1L2Stack,
                                              catsupportbars)
from marxs.missions.arcus.load_csv import load_table
from marxs.missions.arcus.utils import config, id_num_offset
from marxs.missions.arcus.spo import zoom_to_cover_all_spos, rmid_spo


__all__ = [
           'CATfromMechanical',
           'CATWindow',
           'RegularGrid',
           ]


class CATfromMechanical(Parallel):
    '''A collection of diffraction gratings on the Rowland torus.

    After any of the `elem_pos`, `elem_uncertainty` or
    `uncertainty` is changed, `generate_elements` needs to be
    called to regenerate the facets on the GAS.
    '''

    id_col = 'window'

    def stack(self, name):
        return np.vstack([self.data[name + 'X'].data,
                          self.data[name + 'Y'].data,
                          self.data[name + 'Z'].data]).T

    def __init__(self, **kwargs):
        self.channel = kwargs.pop('channel')
        self.conf = kwargs.pop('conf')
        self.data = load_table('gratings', 'facets')

        zoom = [[1, row['xsize'] / 2, row['ysize'] / 2] for row in self.data]
        trans = self.stack('')
        rot = np.stack([self.stack('NormN'),
                        self.stack('DispN'),
                        self.stack('GBarN')])
        pos4d = [compose(trans[i, :], rot[:, i, :].T, zoom[i])
                 for i in range(len(self.data))]

        mirr = np.eye(4)
        if 'm' in self.channel:
            mirr[0, 0] = 1
        else:
            mirr[0, 0] = -1
        if '2' in self.channel:
            mirr[1, 1] = -1

        pos4d = [np.dot(mirr, p) for p in pos4d]

        # Shift pos4d from focal point to center of grating petal
        # Ignore rotations at this point
        pos4d = np.array(pos4d)
        centerpos = np.eye(4)
        centerpos[:, 3] = pos4d.mean(axis=0)[:, 3]
        centerpos_inv = np.linalg.inv(centerpos)
        kwargs['pos4d'] = np.dot(centerpos, kwargs['pos4d'])
        pos4d = np.array([np.dot(centerpos_inv, p) for p in pos4d])

        windowpos = []
        gratingpos = []
        id_start = []
        d_grat = []

        for i in sorted(set(self.data['SPO_MM_num'])):
            ind = self.data['SPO_MM_num'] == i
            winpos = np.eye(4)
            winpos[:, 3] = pos4d[ind, :, :].mean(axis=0)[:, 3]
            winpos_inv = np.linalg.inv(winpos)
            windowpos.append(winpos)
            id_start.append(kwargs.get('id_num_offset', 0) +
                            self.data['facet_num'][ind][0])
            grat_pos = [np.dot(winpos_inv, pos4d[j, :, :])
                        for j in ind.nonzero()[0]]
            gratingpos.append(grat_pos)
            d_grat.append(list(self.data['period'][ind]))

        kwargs['elem_pos'] = windowpos
        kwargs['elem_class'] = CATWindow
        kwargs['elem_args'] = {}
        # This is the elem_args for the second level (the CAT gratings).
        # We need a list of dicts. Each window will then get one dict
        # which itself is a dict of lists
        # currently, 'd' is the only parameter passed down that way
        # but e.g. orderselector could be treated the same way
        kwargs['elem_args']['elem_args'] = [{'d': d,
                                             'order_selector': self.conf['gratinggrid']['elem_args']['order_selector']} for d in d_grat]
        kwargs['elem_args']['elem_pos'] = gratingpos
        kwargs['elem_args']['id_num_offset'] = id_start
        super(CATfromMechanical, self).__init__(**kwargs)

    def generate_elements(self):
        super(CATfromMechanical, self).generate_elements()
        for e in self.elements:
            e.generate_elements()


class CATWindow(Parallel):

    id_col = 'facet'

    def __init__(self, **kwargs):
        kwargs['id_col'] = self.id_col
        kwargs['elem_class'] = CATL1L2Stack
        super().__init__(**kwargs)


class RegularGrid(Sequence):
    '''Regular grid of CAT gratings

    This regular grid of CAT gratings is useful in an early stage of
    development. Gratings are simply put down in a regular grid with
    no thought of mounting structures or optimizing the grating number
    under each SPO. While not the most efficient placement for a real
    mission, this is a useful design to get started before the details
    are worked out.
    '''
    def __init__(self, conf, channels=['1', '2', '1m', '2m'], **kwargs):

        elements = []
        temp, grid_width_y, grid_width_z = zoom_to_cover_all_spos(conf, .5, .5)
        rmid = rmid_spo(conf)

        blazemat = axangle2mat(np.array([0, 0, 1]),
                               np.deg2rad(-conf['blazeang']))
        blazematm = axangle2mat(np.array([0, 0, 1]),
                                np.deg2rad(conf['blazeang']))
        gratinggrid = conf['gratinggrid']

        for chan in channels:
            sig = 1 if '1' in chan else -1
            is_m = -1 if 'm' in chan else +1
            gratinggrid['rowland'] = conf['rowland_' + chan]
            b = blazematm if 'm' in chan else blazemat
            gratinggrid['elem_args']['orientation'] = b
            gratinggrid['normal_spec'] = conf['pos_opt_ax'][chan].copy()
            gratinggrid['guess_distance'] = is_m * (conf['rowland_central'].r + conf['rowland_central'].R)

            y_range = [-grid_width_y, grid_width_y]
            z_range = [sig * (rmid - grid_width_z),
                       sig * (rmid + grid_width_z)]
            z_range.sort()
            elements.append(RectangularGrid(y_range=y_range, z_range=z_range,
                                            id_num_offset=id_num_offset[chan],
                                            **gratinggrid))
        elements.extend([catsupportbars])
        super().__init__(elements=elements, **kwargs)



'''Analyze a grid of monoenergetic simulations for R and Aeff
'''
import os
import glob
import numpy as np
from astropy.table import QTable, Table
import astropy.units as u
from marxs.analysis.gratings import (CaptureResAeff_CCDgaps,
                                     average_R_Aeff)
from marxs.base import check_meta_consistent, check_energy_consistent
from marxs.missions.arcus.utils import id_num_offset
from marxs.missions.arcus.arcus import defaultconf
from marxs.missions.arcus.load_csv import load_table


def zeropos(coord, channel, conf=defaultconf):
    '''Get position of zeroth order

    Parameters
    ----------
    coord : string
        Name of coordinate
    channel : string
        Channel id, e.g. "1m"
    conf : dict
        Arcus configuration dictionary

    Returns
    -------
    pos : float
        coordinate of zeroth order
    '''
    match coord:
        case "circ_phi":
            return np.arcsin((conf['pos_opt_ax'][channel][0]) /
                                      conf['rowland_detector'].r)
        case "circ_y":
            return -conf['pos_opt_ax'][channel][1]
        case "proj_y":
            return conf['pos_opt_ax'][channel][1]
        case "proj_x":
            return conf['pos_opt_ax'][channel][0]
        case _:
            raise NotImplementedError


def analyze_sim(photons, orders, reference_meta, conf):
    '''Get R and Aeff from a single photon event list

    Parameters
    ----------
    photons : `astropy.table.Table`
        Photon event list
    orders : np.array
        list of orders to be analyzed
    reference_meta : dict or None
        If not ``None`` check that the meta information of the photon
        list matches the meta information in ``reference_meta``
    conf : dict
        Arcus configuration (the zero order position for each channel
        is taken from this dict).

    Returns
    -------
    res : np.array
        Array of R values for each aperture and order, measured from
        photons that hit a CCD. For signal close to a CCD boundary, the
        detected photon distribution may be artificially narrow.
        In general it is better to use res (because the CCDs don't follow
        the Rowland circle exactly) but it's so close that ``circ_phi``
        can be used for those cases. ``res`` is set to ``np.nan`` if there
        is no signal (e.g. a chip gap).
    Aeff : quantity
        Effective area for each aperture and order, using only photons
        that are not L1 cross-dispersed and detected on a CCD.
    '''
    n_apertures = len(list(id_num_offset.values()))
    photons['aper_id'] = np.digitize(photons['xou'],
                                     bins=list(id_num_offset.values()))

    if reference_meta is not None:
        check_meta_consistent(photons.meta, reference_meta)
    check_energy_consistent(photons)

    chan_name = list(id_num_offset.keys())

    res = np.zeros((n_apertures, len(orders)))
    aeff = np.zeros_like(res) * u.cm**2


    for a in range(1, n_apertures + 1):
        pa = photons[(photons['aper_id'] == a) &
                     (photons['order_L1'] == 0)]
        resaeff = CaptureResAeff_CCDgaps(A_geom=photons.meta['A_GEOM'] * u.cm**2,
                                         order_col='order',
                                         orders=orders,
                                         dispersion_coord='circ_phi',
                                         zeropos=zeropos('circ_phi',
                                                         chan_name[a - 1],
                                                         conf),
                                         aeff_filter_col='CCD',
                                         )
        out = resaeff(pa, n_photons=int(pa.meta['EXPOSURE']))

        res[a - 1, :] = out['R']
        aeff[a - 1, :] = out['Aeff']
    return res, aeff


def aeffRfromraygrid(inpath, conf,
                     orders=np.arange(-15, 5),
                     allow_inconsistent_meta=False):
    '''Analyze a grid of simulations for R and Aeff

    Parameters
    ----------
    inpath : string
        Path to the simulations grid
    conf : dict
        Arcus configuration (the zero order position for each channel
        is taken from this dict).
    orders : list
        List of integer order numbers to be analyzed
    apertures : list
        List of aperture ids in the simulations
    allow_inconsistent_meta : bool
        If ``False`` (the default) the routine checks that all ray files in
        the grid have compatible meta information (e.g. code version).
        Setting this to ``True`` bypasses that check.

    Returns
    -------
    out : `astropy.table.QTable`
        Table with instrument performance metrics
    '''
    rayfiles = glob.glob(os.path.join(inpath, 'wave*.fits'))
    rayfiles.sort()
    r0 = Table.read(rayfiles[0])
    energies = np.zeros(len(rayfiles)) * u.keV
    # number of apertures is hardcoded
    res = np.zeros((len(rayfiles), 4, len(orders)))
    aeff = np.zeros_like(res) * u.cm**2

    for i, rayfile in enumerate(rayfiles):
        obs = Table.read(rayfile)
        res_i, aeff_i = analyze_sim(obs, orders,
                                    None if allow_inconsistent_meta else r0.meta,
                                    conf)
        res[i, :, :] = res_i
        aeff[i, :, :] = aeff_i
        energies[i] = obs.meta['ENERGY'] * u.keV

    wave = energies.to(u.Angstrom, equivalencies=u.spectral())
    res_4, aeff_4 = average_R_Aeff(res, aeff, axis=1)
    indnon0 = orders != 0
    res_disp, aeff_disp = average_R_Aeff(res_4[:, indnon0], aeff_4[:, indnon0],
                                         axis=1)
    # Columns get filled when writing to fits
    res_4.fill_value = np.nan
    out = QTable([energies, wave,
                  aeff, res,
                  aeff_4, res_4.filled(np.nan),
                  aeff_disp, res_disp],
                 names=['energy', 'wave',
                        'Aeff','R',
                        'Aeff4', 'R4',
                        'Aeff_disp', 'R_disp'])
    out.meta = r0.meta
    for i, o in enumerate(orders):
        out.meta['ORDER_{}'.format(i)] = o
    return out


def orders_from_meta(meta):
    orders = []
    i = 0
    while 'ORDER_{}'.format(i) in meta:
        orders.append(meta['ORDER_{}'.format(i)])
        i += 1
    return np.array(orders)


def csv_per_order(infile, col, outfile):
    '''Rewrite one column in ``aeffRfromraygrid`` to csv file

    Turn one vector-valued (all orders in one cell) column into a
    csv table with one entry per cell.
    '''
    tab = Table.read(infile)
    outtab = Table(tab[col], names=['order_{0}'.format(o) for o
                                    in orders_from_meta(tab.meta)])
    outtab.add_column(tab['wave'], index=0)
    outtab.write(outfile, format='ascii.csv', overwrite=True)


def tg_resolve_events_circdet(evt, defaultconf):
    '''Resolve the tangent plane coordinates for a circular detector

    .. warning::
        This method is still under development.

    Parameters
    ----------
    evt : `astropy.table.Table`
        Table with photon events
    defaultconf : dict
        Arcus configuration dictionary
    '''

    facets = load_table('gratings', 'facets')
    effective_grating_period = np.cos(defaultconf['blazeang'] * u.degree) * np.mean(facets['period']) * facets['period'].unit
    # without the blaze angle would look like this
    # effective_grating_period = np.mean(facets['period']) * facets['period'].unit


    ang_r0 = zeropos('circ_phi', evt.meta['CHANNEL']) * u.rad
    evt['tg_r'] = (-(ang_r0 - evt['circ_phi']) / 2).to(u.degree)
    l = 2 * defaultconf['rowland_central'].r * u.mm * np.cos(ang_r0 - evt['circ_phi'])
    dy = defaultconf['pos_opt_ax'][evt.meta['CHANNEL']][1] * u.mm - evt['circ_y']
    evt['tg_d'] = np.arctan(dy / l).to(u.degree)

    evt['tg_mlam'] = (effective_grating_period * np.sin(evt['tg_r'])).to(u.Angstrom)
